//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// codec.c - functions implementing the various chunk compression/decompression
//	     techniques, plus checksumming
//

#include <stdio.h>
#include <string.h>

#include "common.h"
#include "locodef.h"

// RLE decode the given data and return a malloc'ed decoded copy
// (we don't need to return the length because that's "obvious" from the
// contents of the data)
__declspec(dllexport) int LocoTool_rledecode(uint8_t *data, uint32_t len, uint8_t **outdata, uint32_t *outlen)
{
  uint8_t *chunk;
  uint32_t chunklen, chunkofs;

  chunklen = len*2;
  chunk = my_alloc(chunklen, "chunk");

  chunkofs = 0;
  while (len>0) {
	int8_t rle;
	int run;

	rle = *(int8_t*) data++; len--;
	run = abs(rle)+1;

	if (chunkofs+run > chunklen) {
		uint32_t oldlen = chunklen;
		while (chunkofs+run > chunklen)
			chunklen <<= 1;
		chunk = my_realloc(chunk, chunklen, oldlen, "chunk");
	}

	if (rle<0) {
		uint8_t byte = *data++; len--;

		memset(chunk+chunkofs, byte, run);
	} else {
		memcpy(chunk+chunkofs, data, run);
		len-=run;
		data+=run;
	}
	chunkofs += run;
  }

  *outdata = my_realloc(chunk, chunkofs, chunklen, "chunk");
  *outlen = chunkofs;

  return 1;
}

// RLE decode then string decompression
__declspec(dllexport) int LocoTool_decompress(uint8_t *data, uint32_t len, uint8_t **outdata, uint32_t *outlen)
{
  uint8_t *rle, *chunk;
  uint32_t rlelen, rleofs, chunklen, chunksize;

  if (!LocoTool_rledecode(data, len, &rle, &rlelen))
	return 0;

  chunksize = rlelen*2;
  chunk = my_alloc(chunksize, "chunk");

  chunklen = 0;
  rleofs = 0;
  while (rleofs < rlelen) {
	uint8_t code;
	uint32_t len, ofs;

	code = rle[rleofs++];

	if (chunklen+33 > chunksize) {
		uint32_t oldsize = chunksize;
		while (chunklen+33 > chunksize)
			chunksize <<= 1;
		chunk = my_realloc(chunk, chunksize, oldsize, "chunk");
	}

	if (code == 0xff) {
		chunk[chunklen++] = rle[rleofs++];
	} else {
		len = (code & 7)+1;
		ofs = 32 - (code >> 3);

		memcpy(chunk+chunklen, chunk+chunklen-ofs, len);
		chunklen+=len;
	}
  }

  my_free(rle);

  *outdata = my_realloc(chunk, chunklen, chunksize, "chunk");
  *outlen = chunklen;

  return 1;
}


// define a bit rotate (gcc optimizes this into an i386 ROL instruction)
#define _rotl32(a,n) (((a)<<(n))|((a)>>(32-n)))
#define _rotl8(a,n) (((a)<<(n))|((a)>>(8-n)))
#define _rotr32(a,n) (((a)>>(n))|((a)<<(32-n)))
#define _rotr8(a,n) (((a)>>(n))|((a)<<(8-n)))


// descramble using success bit rotates
__declspec(dllexport) int LocoTool_descramble(uint8_t *data, uint32_t len)
{
  uint8_t bits, d;
  uint32_t ofs;

  bits = 1;
  for (ofs=0; ofs<len; ofs++) {
	data[ofs] = _rotr8(data[ofs], bits);
	bits = (bits+2)&7;
  }

  return 1;
}

// RLE encode the given data, and return a malloc'ed encoded copy plus its
// length in the given parameters
__declspec(dllexport) int LocoTool_rleencode(uint8_t *data, uint32_t len, uint8_t **rleout, uint32_t *rlelen)
{
  uint8_t *out;
  uint32_t start, ofs, outlen, outofs;
  int run;
  uint8_t byte, rle;

  outlen = len*2;
  out = my_alloc(outlen, "out");

  ofs = 0;
  outofs = 0;
  while (len) {
	start = ofs;

	byte = data[ofs++];
	len--;

	run = 1;
	if (len && (byte == data[ofs])) {
		while (len && (byte == data[ofs]) && (run < 127)) {
			run++;
			ofs++;
			len--;
		}
		rle = -(run-1);
		out[outofs++] = rle;
		out[outofs++] = byte;
	} else {
		while (len && (byte != data[ofs]) && (run < 126)) {
			byte = data[ofs];
			run++;
			ofs++;
			len--;
		}
		if (len) {
			run--;
			ofs--;
			len++;
		}
		rle = run-1;
		out[outofs++] = rle;
		memcpy(out+outofs, data+start, run);
		outofs += run;
	}
  }

  *rlelen = outofs;
  *rleout = out;

  return 1;
}

// scramble using success bit rotates
__declspec(dllexport) int LocoTool_scramble(uint8_t *data, uint32_t len)
{
  uint8_t bits, d;
  uint32_t ofs;

  bits = 1;
  for (ofs=0; ofs<len; ofs++) {
	data[ofs] = _rotl8(data[ofs], bits);
	bits = (bits+2)&7;
  }

  return 1;
}

// calculate the checksum of the RLE decoded data
__declspec(dllexport) uint32_t LocoTool_checksum(uint8_t *data, uint32_t len)
{
  uint32_t i, c;

  c = 0xF369A75B;
  if (demofile) c = 0xE369A75B;

  // start with the class, skip the subclass and checksum (obviously)
  c = _rotl32(c ^ data[0], 11);
  for (i=4; i<12; i++)
	c = _rotl32(c ^ data[i], 11);

  // then iterate over all remaining bytes
  for (i=16; i<len; i++)
	c = _rotl32(c ^ data[i], 11);

  return c;
}
