//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// common.c - common functions and definitions
//

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <setjmp.h>

#include "locotool.h"
#include "locodef.h"
#include "common.h"
#include "version.h"

#ifdef __windows
#	include <conio.h>
#endif

// error handler proc to call on errors
LocoTool_err_proc_t err_die;
LocoTool_err_proc err_handler = err_die;

// setjmp/longjmp buffer; if set an error will return using longjmp
jmp_buf *err_return = NULL;

void LocoTool_set_err_handler(LocoTool_err_proc handler)
{
  err_handler = handler;
}

const char *LocoTool_get_version()
{
  return "LocoTool V" VERSTR " - Copyright (C) 2004 by Josef Drexler";
}

int LocoTool_get_vernum()
{
  return VERNUM;
}

void err_die(char *message)
{
  fflush(stdout);
  fflush(stderr);

  fputs(message, stderr);

#ifdef __windows
  fputs("\nPress any key to abort.", stderr);

  getch();
#endif
  fputs("\n", stderr);

  exit(1);
}

void LocoTool_die(const char s[], ...)
{
  char message[1024];

  va_list args;

  va_start(args, s);
  vsnprintf(message, sizeof(message), s, args);
  va_end(args);

  if (!quiet)
	err_handler(message);

  if (err_return) {
	longjmp(*err_return, 1);
  }

  exit(1);
}

//
// Heap functions
//
// Unlike regular malloc et.al., they die in an out-of-memory condition,
// and free and relloc also check that the pointer is a valid memory handle
//
// They do this by maintaining a 16-byte structure at the beginning of each
// memory block, and only returning a pointer past this block
//

uint32_t totalallocated = 0;

typedef struct {
	uint32_t magic;
	char name[8];
	uint32_t len;
} heap_info;

void *my_alloc(size_t len, const char *name)
{
  heap_info *p;

  if (heapdebug) fprintf(stderr, "my_alloc(%d, %s)", len, name);

  p = malloc(len+sizeof(heap_info));
  if (!p) die("Not enough memory for %s, wanted %d", name, len);
  memset(p, 0, len+sizeof(heap_info));
  totalallocated += len+sizeof(heap_info);

  if (heapdebug) fprintf(stderr, " = %p; total=%d\n", p, totalallocated);

  p->magic = 0x2343211;
  strncpy(p->name, name, 7);
  p->len = len;

  return p+1;
}

void *my_realloc(void *p, size_t len, size_t oldlen, const char *name)
{
  heap_info *i = (heap_info *)p-1;

  if (i->magic != 0x2343211)
	die("Trying to realloc invalid pointer!");

  if (i->len != oldlen)
	die("realloc old len %d doesn't match oldlen %d", i->len, oldlen);

  if (heapdebug) fprintf(stderr, "my_realloc(%p, %d, %d, %s) [%s]", ((char*)p)-sizeof(heap_info), len, oldlen, name, i->name);

  totalallocated -= i->len;
  i = realloc(i, len+sizeof(heap_info));
  if (!i) die("Not enough memory for %s, wanted %d", name, len);
  totalallocated += len;

  if (heapdebug) fprintf(stderr, " = %p; total=%d\n", i, totalallocated);

  p = i+1;
  if (oldlen<len) memset(((char*)p)+oldlen, 0, len-oldlen);

  i->magic = 0x2343211;
  strncpy(i->name, name, 7);
  i->len = len;

  return p;
}

char *my_strdup(const char *s, const char *name)
{
  char *c = my_alloc(strlen(s)+1,"name");
  strcpy(c, s);
  return c;
}

// this is normally used via the my_free macro which also zeroes p afterwards
void *_my_free(void *p)
{
  if (p) {
	heap_info *i = (heap_info *)p-1;

	if (i->magic != 0x2343211)
		die("Trying to free invalid pointer!");

	totalallocated -= i->len+sizeof(heap_info);

	if (heapdebug) fprintf(stderr, "my_free(%p) [%s] total=%d \n", i, i->name, totalallocated);

	free(i);
  }
  return NULL;
}


// read an unsigned value of the given size
uint32_t getuvalue(uint8_t *data, int size)
{
  uint32_t value = *(uint32_t*)(data);
  value &= 0xffffffff >> (32-8*size);

  return value;
}

// same but signed
int32_t getsvalue(uint8_t *data, int size)
{
  int32_t value = *(int32_t*)(data);
  value = (value << (32-8*size)) >> (32-8*size);

  return value;
}

// and automatic, unsigned for size>0 and signed for size<0
int32_t getvalue(uint8_t *data, int size)
{
  if (size > 0)
	return getuvalue(data, size);
  else
	return getsvalue(data, -size);
}


// get size of varinf structure
uint32_t structsize(varinf *vars)
{
  int i;
  uint32_t ofs;

  ofs = 0;
  for (i=0; vars[i].ofs != -1; i++) {
	if (ofs != vars[i].ofs) die("Structure is invalid, ofs=%X but next field is %X", ofs, vars[i].ofs);
	ofs += abs(vars[i].size) * vars[i].num;
  }
  return ofs;
}
