//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// dump.c - dump object data structure entries into an xml file
//

#include <string.h>
#include <sys/stat.h>

#include "common.h"
#include "codec.h"
#include "options.h"
#include "xml.h"
#include "png.h"

// up to 10 indentations with no silly loop to make tabs
char indent[11] = "\t\t\t\t\t\t\t\t\t\t";
#define INDENT (indent+10-indentation)

//
// ** dump handlers
//
// These dump various variables/structures/bit fields to the xml file
// They all return the number of bytes they ate
//

// write a bit field
uint32_t dumpflags(FILE *xml, uint8_t *data, const char **flags, int size, int indentation)
{
  uint32_t value, i, state;
  const char *name;
  char defname[32];

  if (size > 4) die("Can't dump flags with %d bytes", size);

  value = getvalue(data, size);

  for (i=0; i<size*8; i++) {
	if (*flags) {
		name = *flags;
		flags++;
	} else
		name = "";

	if (!strlen(name)) {
		sprintf(defname, "bit_%X", i);
		name = defname;
	}
	state = (value & (1 << i)) >> i;
	if ((!onlysetbits && (name != defname)) || state)
		fprintf(xml, "%s<bit name=\"%s\">%d</bit>\n", INDENT, name, state);
  }
  return size;
}

// dump a structure to the xml file, structure definition in the vars parameter
uint32_t dodump(FILE *xml, uint8_t *data, varinf *vars, int indentation)
{
  int i, j, size;
  uint32_t dumped, totaldumped, ofs;
  int32_t value;
  const char *name, *fname, *tag;
  char defname[32], arrname[32];

  totaldumped = 0;
  ofs = 0;

  for (i=0; vars[i].ofs != -1; i++) {
	fname = vars[i].name;
	if (!strlen(fname)) {
		sprintf(defname, "field_%X", vars[i].ofs);
		fname=defname;
	}
	if (ofs != vars[i].ofs) die("Structure is invalid, ofs=%X but next field is %X, %s", ofs, vars[i].ofs, fname);
	for (j=0; j<vars[i].num; j++) {
		name = fname;
		if (vars[i].num > 1) {
			sprintf(arrname, "%s[%d]", fname, j);
			name = arrname;
		}

		size = vars[i].size;
		if (!size) {
			FILE *nul;

			if (vars[i].flags || !vars[i].structvars)
				die("Variable %s has no size", name);

			if (vars[i].num > 1)
				die("Variable sized array members not supported");

			nul = fopen("NUL", "wt");
			if (!nul) die("Can't write to NUL: %s", strerror(errno));
			size = dodump(nul, data+ofs, vars[i].structvars, 0);
			fclose(nul);
		}

		if (vars[i].flags) {
			fprintf(xml, "%s<bitmask name=\"%s\" size=\"%d\">\n", INDENT, name, size);
			dumped = dumpflags(xml, data+ofs, vars[i].flags, vars[i].size, indentation+1);
			fprintf(xml, "%s</bitmask>\n", INDENT);
		} else if (vars[i].structvars) {
			fprintf(xml, "%s<structure name=\"%s\" size=\"%d\">\n", INDENT, name, size);
			dumped = dodump(xml, data+ofs, vars[i].structvars,
				indentation+1);

			fprintf(xml, "%s</structure>\n", INDENT);
		} else {
			value = getvalue(data+ofs, size);
			size = abs(size);
			tag = "variable";
			if (!strlen(vars[i].name)) {
				if (value || allunknown)
					tag = "unknown";
				else
					tag = NULL;	// don't print unset unknown variables
			}

			if (tag)
				fprintf(xml, "%s<%s name=\"%s\" size=\"%X\">%d</%s>\n", INDENT, tag, name, size, value, tag);
			dumped = size;
		}
		ofs += dumped;
		totaldumped += dumped;
	}
  }
  return totaldumped;
}

// this is the function actually called by the dump procs, all it does
// is dump the structure in the vars parameter with an initial indentation of 1
uint32_t dumpvars(FILE *xml, uint8_t *data, varinf *vars, int indentation)
{
  return dodump(xml, data, vars, indentation);
}

// dump the language/description info
// those are a bunch of zero-terminated strings preceeded by the language byte
uint32_t dumplang(FILE *xml, uint8_t *data, int num)
{
  int language;
  uint32_t ofs;

  ofs = 0;
  while ( (language = data[ofs++]) != 0xff ) {
	fprintf(xml, "\t<description num=\"%d\" language=\"%d\">", num, language);
	putxmlstr(xml, data+ofs, 0);
	fputs("</description>\n", xml);
	ofs += strlen(data+ofs) + 1;
  }

  return ofs;
}

// dump an object dependence (i.e. an outside object that this object depends on)
uint32_t dumpuseobj(FILE *xml, uint8_t *data, int num, int total, useobjid type, int32_t *classes)
{
  const char *typename;
  char arrtype[32];
  int i, classok;

  typename = useobjnames[type];
  if (total > 1) {
	sprintf(arrtype, "%s[%d]", typename, num);
	typename = arrtype;
  }

  classok = 0;
  for (i=0; classes[i] != -1; i++)
	if ( (classes[i] == data[0]) || (classes[i] == 0x100) )
		classok = 1;

  if (!classok)
	die("Not correct class for %s: %02X", typename, data[0]);

  fprintf(xml, "\t<useobject desc=\"%s\" class=\"%d\">%.8s</useobject>\n",
	typename, data[0], data+4);

  return 16;
}

// dump a cargo capacity entry with various cargo types
uint32_t dumpcap(FILE *xml, uint8_t *data, int num, int total)
{
  uint32_t dumped;
  int capacity, cargotype, refcap;

  if (total);

  capacity = data[0];

  dumped = 1;

  if (capacity) {
	fprintf(xml, "\t<cargo num=\"%d\" capacity=\"%d\">\n", num, capacity);
	while (getvalue(data+dumped, 2) != 0xffff) {
		cargotype = data[dumped++];

		refcap = getvalue(data+dumped, 2);
		dumped += 2;

		fprintf(xml, "\t\t<cargotype id=\"%d\">%d</cargotype>\n", cargotype, refcap);
	}
	dumped += 2;
	fputs("\t</cargo>\n", xml);
  }

  return dumped;
}

// dump the list of sprites plus their contents
uint32_t dumpsprites(FILE *xml, uint8_t *data)
{
  uint32_t i, j, num, size, totalsize, dumped;
  uint8_t *spritedata;
  FILE *png = NULL;
  char *pngname = NULL;

  num = getvalue(data, 4);
  size = getvalue(data+4, 4);
  dumped = 8;

  spritedata = data+8+16*num;
  totalsize = 0;

  if (allinone)
	pngname = makefilename(".png");

  for (i=0; i<num; i++) {
	uint32_t ofs,width,height,flags, spritesize;
	int32_t xofs,yofs;

	ofs=getvalue(data+dumped,4);  dumped+=4;
	width=getvalue(data+dumped,2);  dumped+=2;
	height=getvalue(data+dumped,2);  dumped+=2;
	xofs=getsvalue(data+dumped,2);  dumped+=2;
	yofs=getsvalue(data+dumped,2);  dumped+=2;
	flags=getvalue(data+dumped,4);  dumped+=4;

	fprintf(xml, "\t<sprite id=\"%d\" xofs=\"%d\" yofs=\"%d\">\n", i, xofs, yofs);
	dumpflags(xml, (uint8_t*) &flags, spriteflags, 4, 2);

	if ( (width==1) && (height==1) ) {
		uint8_t col;

		spritesize = putspriterow(spritedata+ofs, flags, 0, width, &col);

		fprintf(xml, "\t\t<stub>%d</stub>\n", col);
	} else if (flags & 0x40) {
		spritesize = 0;
	} else if (embedpng) {
		fputs("\t\t<pngdata>", xml);
		fputs("</pngdata>\n", xml);
		spritesize = 0;
	} else if (allinone) {
		int x, y;

		spritesize = addsprite(spritedata+ofs, i, width, height, flags, &x, &y);
		fprintf(xml, "\t\t<pngimage xofs=\"%d\" yofs=\"%d\" width=\"%d\" height=\"%d\">", x, y, width, height);
		putxmlstr(xml, pngname, 0);
		fputs("</pngimage>\n", xml);
	} else {
		pngname = makepngname(i);
		png = fopen(pngname, "wb");
		if (!png) die("Can't write %s: %s", pngname, strerror(errno));

		fputs("\t\t<pngfile>", xml);
		putxmlstr(xml, pngname, 0);
		fputs("</pngfile>\n", xml);

		my_free(pngname);

		spritesize = makepng(xml, png, spritedata + ofs, width, height, flags);
		fclose(png);

	}

	totalsize += spritesize;

	fputs("\t</sprite>\n", xml);
  }

  dumped += totalsize;
  if (totalsize != size) die("Sprite data different from size; wanted %d had %d", size, totalsize);

  if (allinone) {
	png = fopen(pngname, "wb");
	if (!png) die("Can't write %s: %s", pngname, strerror(errno));
	makepng(xml, png, NULL, 0, 0, 0);
	fclose(png);
	my_free(pngname);
  }

  return dumped;
}

uint32_t dumpaux(FILE *xml, uint8_t *data, int class, int nameind, int id, int numid, int size, int num)
{
  int i, type, siz;
  uint32_t dumped;
  char arrname[32], basename[32];
  const char *name, *auxname;
  varinf *vars;
  varinf auxvars[] = {
	{ 0x00, 0, 0, "" },
	{ -1 },
  };

  sprintf(basename, "aux_%d", nameind);

  auxname = objclasses[class].aux[nameind].name;
  if (!strlen(auxname))
	auxname = basename;


  if (numid) {
	snprintf(arrname, sizeof(arrname)-1, "%s[%d]", auxname, id);
	name = arrname;
  } else
	name = auxname;

  type = 0;
  dumped = 0;
  siz = abs(size);
  if (num < 0) {
	type = -num;
	num = 0;
	for (i=0; getsvalue(data+i, type) != -1; i+=siz) num++;
	dumped+=type;
  }

  fprintf(xml, "\t<auxdata name=\"%s\" size=\"%d\" num=\"%d\" type=\"%d\">\n", name, siz, num, type);

  if (size < 0) {
	num *= -size;
	size = 1;
  }
  auxvars[0].size = size;

  vars = objclasses[class].aux[nameind].vars;
  if (!vars) vars = auxvars;

  siz = structsize(vars);
  if (siz != num * size) {
	vars[0].num = 1;
	siz = structsize(vars);
	vars[0].num = num * size / siz;
	if (vars[0].num * siz != size * num)
		die("%s size %d*%d != %d*%d", name, siz, vars[0].num, size, num);
  }

  dumped += dumpvars(xml, data, vars, 2);

  fputs("\t</auxdata>\n", xml);

  return dumped;
}

uint32_t dumpstrtable(FILE *xml, uint8_t *data, int id, int num)
{
  int i;
  uint32_t dumped;

  dumped = num*2;
  fprintf(xml, "\t<stringtable id=\"%d\" num=\"%d\">\n", id, num);
  for (i=0; i<num; i++) {
	dumped = getvalue(data+i*2, 2);
	fprintf(xml, "\t\t<string id=\"%d\" type=\"%d\">", i, data[dumped+strlen(data+dumped)+1]);
	dumped += putxmlstr(xml, data+dumped, 0)+2;
	fputs("</string>\n", xml);
  }
  fputs("\t</stringtable>\n", xml);

  return dumped;
}

uint32_t dumpsounds(FILE *xml, uint8_t *data)
{
  FILE *wav;
  char *wavname;
  uint32_t len, wavlen, dumped, extra;
  int i;

  dumped = 0;

  if (getvalue(data+dumped, 4) != 1)
	die("Unsupported number of samples");
  dumped += 4;

  len=getvalue(data+dumped, 4);
  dumped += 4;

  for (i=0; i<4; i++) {
	if (getvalue(data+dumped, 4))
		die("Unsupported aux data in sample header");
	dumped += 4;
  }

  if (getvalue(data+dumped, 4) != 1)
	die("Unsupported number of samples");
  dumped += 4;

  if (getvalue(data+dumped, 4) != 8)
	die("Unsupported sample header size");
  dumped += 4;

  wavlen = getvalue(data+dumped, 4);
  dumped += 4;

  // extra bytes which I do not understand
  extra = len - (16+8+4+wavlen);
  wavlen += extra;

  wavname = makefilename(".wav");
  wav = fopen(wavname, "wb");
  if (!wav) die("Cannot write %s: %s", wavname, strerror(errno));

  fprintf(xml, "\t<wavfile size=\"%d\">%s</wavfile>\n", extra, wavname);
  my_free(wavname);

  fwrite("RIFF", 1, 4, wav);

  len = 8+4+16+4+4+wavlen;	// "WAVEfmt "+len+header+"data"+len+wavedata
  fwrite(&len, 4, 1, wav);

  fwrite("WAVEfmt ", 1, 8, wav);

  len = 16;
  fwrite(&len, 4, 1, wav);
  fwrite(data+dumped, 1, len, wav);
  dumped += len;

  fwrite("data", 1, 4, wav);

  fwrite(&wavlen, 4, 1, wav);
  fwrite(data+dumped, 1, wavlen, wav);
  dumped += wavlen;

  fclose(wav);

  return dumped;
}

//
// *** "main" dump functions
//

#ifdef DEBUGDMP
FILE *dmp;
extern int colhist[256];
#endif

// dump the object class-specific data
int dumpchunk(FILE *xml, uint8_t *data, uint32_t len, int class)
{
  uint32_t dumped;
  int i, j, num, num2;
  int32_t *params;

#ifdef DEBUGDMP
  fwrite(data, 1, len, dmp);
  fflush(dmp);
#endif

  if ( (class >= NUMOBJCLASSES) || !objclasses[class].desc)
	die("Can't translate class %02X yet", class);

  dumped = 0;
  if (objclasses[class].desc[0].type != desc_objdata)
	die("Object class has no objdata at start");

  for (i=0; objclasses[class].desc[i].type != desc_END; i++) {
	params = objclasses[class].desc[i].param;

	switch (objclasses[class].desc[i].type) {
		case desc_objdata:	// -
			dumped += dumpvars(xml, data+dumped, objclasses[class].vars, 1);
			break;

		case desc_lang:		// langid
			dumped += dumplang(xml, data+dumped, params[0]);
                        break;

		case desc_useobj:	// num* type classes...
			for (j=0; ; j++) {
				num = getnum(data, &dumped, params[0]);

				if (loopescape(j,num)) break;

				dumped += dumpuseobj(xml, data+dumped, j, num, params[1], params+2);
			};
			break;

		case desc_auxdata:	// nameind numaux* size num*
			num = getnum(data, &dumped, params[1]);
			num2 = getnum(data, &dumped, params[3]);
			for (j=0; !loopescape(j,num); j++)
				dumped += dumpaux(xml, data+dumped, class, params[0], j, num, params[2], num2);
			break;

		case desc_auxdatafix:	// nameind numaux* size numsize
			num = getnum(data, &dumped, params[1]);
			for (j=0; !loopescape(j,num); j++) {
				num2 = getvalue(data+dumped, params[3]);
				dumped += params[3];

				dumped += dumpaux(xml, data+dumped, class, params[0], j, num, params[2], num2);
			}
			break;

		case desc_auxdatavar:	// nameind numaux* size type
			num = getnum(data, &dumped, params[1]);
			for (j=0; !loopescape(j,num); j++)
				dumped += dumpaux(xml, data+dumped, class, params[0], j, num, params[2], -params[3]);
			break;

		case desc_strtable:	// id num* ofs*
			dumped += dumpstrtable(xml, data+dumped, params[0], data[params[1]]);
			break;

		case desc_cargo:	// num*
			num = getnum(data, &dumped, params[0]);
			for (j=0; !loopescape(j,num); j++)
				dumped += dumpcap(xml, data+dumped, j, num);
			break;

		case desc_sprites:	// -
			dumped += dumpsprites(xml, data+dumped);
			break;

		case desc_sounds:	// -
			dumped += dumpsounds(xml, data+dumped);
			break;

		default:
			die("Unknown object description type %d", objclasses[class].desc[i].type);
	}
  }

  return dumped;
}

// Exported: decode the given file into an xml an a slew of png files
int LocoTool_decode(char *datname)
{
  int class;
  uint32_t subclass;
  uint8_t *chunk;
  uint32_t dumped;

  uint8_t *data, *ptr, *end;
  struct stat st;
  FILE *dat, *xml;
  char *xmlname;
  jmp_buf decode_err;

  // in case of an error, we return here with 0
  err_return = &decode_err;
  if (setjmp(*err_return)) {
	err_return = NULL;
	return 0;
  }

  setfilebase(datname);

  xmlname = makefilename(".xml");

  dat = fopen(datname, "rb");
  if (!dat) die("Can't read %s: %s", datname, strerror(errno));

  // get the file size
  if (fstat(fileno(dat), &st)) die("Can't stat %s: %s", datname, strerror(errno));

  // slurp the whole file
  data = my_alloc(st.st_size, "data");
  fread(data, 1, st.st_size, dat);
  fclose(dat);

  // sanity check that it's really a Locomotion .dat file THIS IS NOT CORRECT
  //if (data[3] != 0x11) die("File %s is not a .dat file", datname);
  end = data+st.st_size;

  class = data[0] & 0x3F;
  subclass = getvalue(data+1, 3);

  if ( ((class & 0x3F) >= NUMOBJCLASSES) || !objclasses[class&0x3f].desc)
	die("Can't translate class %02X yet", class);

  if (data[16] > 3)
	die("Can't decode chunk compression %d", data[16]);

  xml = fopen(xmlname, "wt");
  if (!xml) die("Can't write %s: %s", xmlname, strerror(errno));
  my_free(xmlname);

  // dump the file contents as xml
  fputs("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n", xml);

  fprintf(xml, "<object class=\"0x%02X\" subclass=\"0x%06X\" name=\"",
	data[0], subclass);
  putxmlstr(xml, data+4, 8);
  fputs("\">", xml);

  ptr = data+16;

  class &= 0x3f;

  // For debugging the object data assembly functions
#ifdef DEBUGDMP
  {
    char *dmpname = makefilename(".dmp");
    dmp=fopen(dmpname,"wb"); if (!dmp) die("write %s: %s", dmpname, strerror(errno));
    my_free(dmpname);
    fwrite(data, 1, 16, dmp);
#endif

  while (ptr < end) {
	int compression;
	uint32_t len, chunklen;

	compression = *ptr; ptr++;
	len = *(uint32_t*) ptr; ptr+=4;

	fprintf(xml, "<chunk compression=\"%d\">\n", compression);
	switch(compression) {
		case 0:
			dumpchunk(xml, ptr, len, class);
			break;
		case 1:
			if (!LocoTool_rledecode(ptr, len, &chunk, &chunklen))
				die("Error while RLE decoding");
			dumpchunk(xml, chunk, chunklen, class);
			my_free(chunk);
			break;
		case 2:
			if (!LocoTool_decompress(ptr, len, &chunk, &chunklen))
				die("Error while RLE decompressing");
			dumpchunk(xml, chunk, chunklen, class);
			break;
		case 3:
			LocoTool_descramble(ptr, len);
			dumpchunk(xml, ptr, len, class);
			break;
		default:
			die("Can't decode chunk compression %d", compression);
	}

	fputs("</chunk>", xml);
	ptr += len;
  }
  fputs("</object>\n", xml);

#ifdef DEBUGDMP
    fclose(dmp);
/*
    dmpname = makefilename(".col");
    dmp=fopen(dmpname,"wb"); if (!dmp) die("write %s: %s", dmpname, strerror(errno));
    my_free(dmpname);
    fwrite(colhist, 1, sizeof(colhist), dmp);
    fclose(dmp);
*/
      }
#endif

  fclose(xml);

  my_free(data);

  clearoptions();

  // if an error happens later, we *don't* want to return to the above place!
  err_return = NULL;
  return 1;
}
