#ifndef _LOCODEF_H
#define _LOCODEF_H

//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// locodef.h - data structures and variables used across the encoding/decoding
//	       functions
//

#include <stdio.h>
#include <stdint.h>
#include <png.h>

// Global variables

extern int embedpng;	// embed PNG data in the XML file
extern int allinone;	// put all sprites into a single PNG file
extern int palettetype;	// which colour palette to use for company colours
extern int quiet;	// no output, no messages, fail silently
extern int demofile;	// if 1, generate checksum for the demo
extern int heapdebug;	// to debug the heap functions


#include "locotool.h"


// sprite flag definition; used by various handlers

extern const char *spriteflags[];


// enum of all supported XML elements
//
// el_document is pseudo-element of the notional document that contains
// the xml data of the file
//
// NOTE: Any changes here must be synchronized with the elemlist[] array in xml.c
//
enum en_elemid {
	el_bit, el_bitmask, el_cargo, el_cargotype, el_chunk,
	el_description, el_object, el_pngfile, el_sprite,
	el_structure, el_unknown, el_useobject, el_variable,
	el_stub, el_auxdata, el_stringtable, el_string,
        el_wavfile, el_pngimage,

	el_document,
};
typedef enum en_elemid elemid;


// enum of all supported XML attributes
//
// the first set have a number as argument, the second set after at_STRINGVALUE
// have a string as argument
//
// NOTE: Any changes here must be synchronized with the attname[] array in xml.c
//
enum en_attrid {
	at_capacity, at_class, at_compression, at_id, at_type,
	at_language, at_num, at_size, at_subclass, at_xofs, at_yofs,
	at_width, at_height,

	at_STRINGVALUE,

	at_name, at_desc,

	at_LASTVALUE,
};
typedef enum en_attrid attrid;

extern const char *attname[at_LASTVALUE];


// enum of all support <useobject> types
//
// NOTE: Any changes here must be synchronized with the useobjnames[] array in xml.c
//
enum en_useobjid {
	ob_tracktype, ob_trackmod, ob_visualeffect, ob_wakeeffect,
	ob_rackrail, ob_compatible, ob_startsnd, ob_soundeffect,
	ob_signal, ob_tunnel, ob_bridge, ob_station, ob_cliff,
	ob_roadmod, ob_produces, ob_accepts, ob_fence, ob_cargo,
	ob_default,

	ob_LASTVALUE,
};
typedef enum en_useobjid useobjid;

extern const char *useobjnames[ob_LASTVALUE];


//
// definition of an object structure variable
//
struct s_varinf {
	int ofs;		// offset counted from the beginning of the structure
	int size;		// size in bytes, negative for signed variables
	int num;		// number of entries (for arrays)
	const char *name;	// name, if known; if empty (*not* NULL!) will use generic field_## name
	struct s_varinf *structvars;	// if not NULL, sub-structure definition
	const char **flags;		// if not NULL, bit field definition (pointer to list of *char bit names)
};
typedef struct s_varinf varinf;

//
// definition of a sprite
//
struct s_spriteinf {
	uint32_t width,height,flags, spritesize;
	int32_t xofs,yofs;
};
typedef struct s_spriteinf spriteinf;

//
// definition of a description/language text
//
struct s_descinf {
	int num, lang;
	uint32_t len;
	uint8_t data[1];
};
typedef struct s_descinf descinf;

//
// definition of an external object dependence
//
struct s_useobjinf {
	int class, num;
	useobjid type;
	uint8_t obj[16];	// the binary code in the .dat file
};
typedef struct s_useobjinf useobjinf;

//
// definition of auxiliary data
//
struct s_auxdatainf {
	int nameind, id, num, size, type;
	uint32_t len;
	uint8_t data[1];
};
typedef struct s_auxdatainf auxdatainf;

//
// string table entry
//
struct s_string {
	int id, num, type;
	uint32_t len;
	char str[1];
};
typedef struct s_string string;

//
// attribute value; either a number or string depending on whether it
// is defined before or after at_STRINGVAL in en_attrid
//
union u_attval {
	uint32_t u32;
	char *str;
};
typedef union u_attval attval;

//
// description of an element; can contain either a list of bit names or
// a sub-structure, depending on the element
//
union u_elemdesc {
	const char **flags;
	varinf *vars;
};
typedef union u_elemdesc elemdesc;

//
// information about an XML element
//
struct s_eleminf {
	elemid id;			// element ID
	attval atts[at_LASTVALUE];	// attribute values
	elemdesc desc;			// description of contents
	void *data;			// pointer to corresponding data in object structure
};
typedef struct s_eleminf eleminf;

//
// definition about an array holding XML information
// these arrays can be automatically made longer using the checklen() function
//
struct s_arrayinf {
		// size: number of bytes allocated to the array (may grow)
		// ent:  number of entries defined (may grow or become smaller)
		// entsize: number of bytes per entry (set at initialization)
	uint32_t size, ent, entsize;

		// pointer to array of data entries
		// e.g. for useobj_i this points to useobj
		// (set at initialization)
	char ***data;
};
typedef struct s_arrayinf arrayinf;

//
// all information collected from the XML file
//
struct s_xmlinf {

	//
	// META DATA (only used while parsing, doesn't contain object data)
	//

		// does the current element need character data?
	int chardata;

		// current object character data, if chardata is non-zero
		// (i.e. the stuff betyween <element> and </element>)
	uint32_t curvaluelen, curvaluesize;
	char *curvalue;

	char *curpngfile;
	png_byte **curpngrows;
        uint32_t curpngwidth, curpngheight;

		// current element stack
	arrayinf elems_i;
	eleminf **elems;

	//
	// OBJECT DATA
	//

		// object class, chunk compression
	int class, compression;

		// object subclass (no idea what this means)
	uint32_t subclass;

		// object name, space padded to 8 bytes
	uint8_t name[8];

		// object data structure; size is object class specific
	arrayinf objdata_i;
	uint8_t *objdata;

		// language/description data
	arrayinf descdata_i;
	descinf **descdata;

		// external object dependencies
	arrayinf useobj_i;
	useobjinf **useobj;

		// capacity data for both cargo bays
	uint32_t capsize[2], caplen[2];		// data size and used length
	uint8_t *caps[2];			// capacity data

		// auxiliary data
	int nextauxid;
	arrayinf auxdata_i;
	auxdatainf **auxdata;

		// string tables
	arrayinf strtable_i;
	string **strtable;
	int curstrtableid;

		// list of sprite information (width/height/etc)
	arrayinf sprites_i;
	spriteinf **sprites;

		// sprite data
	uint32_t cursprite;
	arrayinf spritedata_i;
	uint8_t **spritedata;
};
typedef struct s_xmlinf xmlinf;

/*
//
// definition of the handlers to dump to xml or parse xml into the object data
//
typedef uint32_t dumpproc(FILE*, uint8_t*, uint32_t);
typedef void parseproc(xmlinf*);
*/

//
// definition/description of an object information item
//
enum en_desctype {
	// type			   argument(s)
	desc_objdata,		// -
	desc_lang,		// langid
	desc_useobj,		// num* type classes...
	desc_auxdata,		// nameind numaux* size num* (regular auxdata, known number of entries)
	desc_auxdatafix,	// nameind numaux* size numsize (fixed length auxdata, length stored before auxdata)
	desc_auxdatavar,	// nameind numaux* size type (variable length auxdata, delimited by 0xFF*type)
	desc_strtable,		// id numofs ofsofs
	desc_cargo,		// num*
	desc_sprites,		// -
	desc_sounds,		// -

	// * num entries have these possibilities:
	//	num > 0 means fixed number (array support)
	//	num = 0 means fixed number, single entry (no array support)
	//	num < 0: num is an index into the objdata array where the
	//		 actual number of entries is stored (array support)
        //
	//	num + 0x01xx0000 mean until data[] == 0xff; only supported for desc_useobj
	//	num + 0x02xx0000 means use data[num] & xx
	//	num + 0x03000000 means use data[num] != 0 ? 1 : 0
	//	num + 0x04000000 means special calculation type in calcdescnum[] in objects.c

	desc_END,
};
typedef enum en_desctype desctype;

#define descnumuntil()  (0x01000000                                   )
#define descnumand(x,y) (0x02000000 | (((y)&0xff)<<16) | ((-x)&0xffff))
#define descnumif(x)    (0x03000000 |                    ((-x)&0xffff))
#define descnumspec(x)  (0x04000000 |                    (( x)&0xffff))

// code to test whether j is in the range allowed by num
// num < 0: no range
// num == 0: 0..0
// num > 0: 0..num-1
#define loopescape(j,num) \
	( (num < 0) || ( (num == 0) && (j > 0) ) || ( (num > 0) && (j >= num) ) )

int getnum(uint8_t *data, uint32_t *ofs, int32_t numdef);

struct s_objdesc {
	desctype type;
	int32_t param[5];
};
typedef struct s_objdesc objdesc;

struct s_auxdesc {
	const char *name;
	varinf *vars;
};
typedef struct s_auxdesc auxdesc;


//
// object class-specific definitions and handlers
//
struct s_objclass {
	varinf *vars;
	uint32_t size;
	auxdesc *aux;
	objdesc *desc;
};
typedef struct s_objclass objclass;

//
// array object class-specific definitions and handlers for each class
//
#define NUMOBJCLASSES 0x22
extern objclass objclasses[NUMOBJCLASSES];

// name alias list
extern const char* namealiases[][2];

#endif // ndef _LOCODEF_H
