//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// objects.c - object class-specific data structure definitions
//

#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "locodef.h"

//
// Two main contents of this file:
// - the object class-specific data structures
//   (in varinf arrays plsu char* arrays for the bit fields)
// - the object class-specific item descriptions
//
// Each varinf entry is either a simple variable, a sub-structure, a
// bit field, or an array of those.  All must be in sequence without
// holes.
//
// The descriptions tell LocoTool what kind of data is arranged in which
// order in the .dat file.
//
//
// All object classes are then put into the objclasses[] array at the bottom
// of this file.
//

// ***********************
//      General items
// ***********************

	// the known bits in the flags of each sprite
const char *spriteflags[] = {
	"hasdata", "", "chunked", "",
	"", "", "copy", "",
	NULL
};

	// if a variable etc. can have different names, enter the aliases here
const char* namealiases[][2] = {
//	{ "oldname", "newname" },
	{ "wideslope", "steepslope" },
	{ "curvetypes", "trackpieces" },
	{ "salecostfact", "sellcostfact" },
	{ "numdata", "numaux01" },
	{ "numaux2ent ", "numtiles" },
	{ "numaux4", "numnodes" },
	{ "numaux5", "numedges" },
	{ "approach", "flight" },
	{ "helitakeoff", "helibegin" },
	{ "heliland", "heliend" },
	{ "takeoff", "touchdown" },
	{ NULL },
};

	// for bits whose meaning is not predefined
const char *nobits[] = { NULL };

// sometimes the number of entries of an item is not simply constant or
// stored somewhere; in this case these handlers return the calculated number
// see also the comment in en_desctype in locodef.h
int vehnumtrack(uint8_t *data)
{
  return (data[2] < 2) && !(data[0xE1] & 2) ? 0 : -1;
}

int vehnumrackrail(uint8_t *data)
{
  return (data[2] < 2) && (data[0xE0]&0x40) ? 0 : -1;
}

typedef int calcdescnumproc(uint8_t *);

calcdescnumproc *calcdescnum[] = {
	vehnumtrack, vehnumrackrail,
};

int getnum(uint8_t *data, uint32_t *ofs, int32_t numdef)
{
  int8_t type, arg;
  int16_t num;

  type = numdef >> 24;
  arg = (numdef & 0xff0000) >> 16;
  num = numdef & 0xffff;

  switch (type) {
	case 0:
		return num;

	case 1:
		if (data[*ofs] != 0xff)
			return 0xffffff;

		(*ofs)++;
		return -1;

	case 2:
		num = data[-num] & arg;
		if (num)
			return num;
		else
			return -1;

	case 3:
		return data[-num] ? 0 : -1;

	case 4:
		return calcdescnum[num](data);

	case -1:
		if (num < 0) {
			num = data[-num];
			if (num)
				return num;
			else
				return -1;
		}

		// else fall through

  }

  die("Invalid description count %x", numdef);
}


// ***********************
//  SIMPLE CLASS HANDLER
// ***********************

varinf simplevars[] = {
	{ 0x00, 1, 6, "" },
	{ -1 },
};

objdesc simpledesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 00:  INTERFACES
// ***********************

varinf interfacevars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 4, 1, "imageId" },
	{ 0x06, 1, 18, "colours" },
	{ -1 },
};


// ***********************
// Class 01:  SOUNDS
// ***********************

varinf sfxvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 4, 1, "data_ptr" },
	{ 0x06, 1, 1, "var_06" },
	{ 0x07, 1, 1, "" },
	{ 0x08,-4, 1, "volume" },
	{ -1 },
};

objdesc sfxdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_sounds },
	{ desc_END },
};


// ***********************
// Class 02:  CURRENCIES
// ***********************

varinf currvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 1, "prefixSymbolId" },
	{ 0x04, 2, 1, "suffixSymbolId" },
	{ 0x06, 4, 1, "objectIconId" },
	{ 0x0A, 1, 1, "seperator" },
	{ 0x0B, 1, 1, "factor" },
	{ -1 },
};

objdesc currdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_lang, { 1 } },
	{ desc_lang, { 2 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 03:  EXHAUST FX
// ***********************

varinf exhfxvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 1, "numImages" },
	{ 0x04, 1, 1, "numStationaryTicks" },
	{ 0x05, 1, 1, "spriteWidth" },
	{ 0x06, 1, 1, "spriteHeightNegative" },
	{ 0x07, 1, 1, "spriteHeightPositive" },
	{ 0x08, 2, 1, "flags" },
	{ 0x0A, 4, 1, "var_0A" },
	{ 0x0E, 4, 1, "baseImageId" },
	{ 0x12, 2, 1, "totalNumFramesType0" },
	{ 0x14, 2, 1, "totalNumFramesType1" },
	{ 0x16, 4, 1, "frameInfoType0_ptr" },
	{ 0x1A, 4, 1, "frameInfoType1_ptr" },
	{ 0x1E, 1, 1, "numSoundEffects" },
	{ 0x1F, 1, 9, "soundEffectIds" },
	{ -1 },
};

auxdesc exhfxaux[2] = {
	{ "", NULL },
	{ "", NULL },
};

objdesc exhfxdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_auxdatavar, { 0, 0, 2, 1 } },
	{ desc_auxdatavar, { 1, 0, 2, 1 } },
	{ desc_useobj, { -0x1e, ob_soundeffect, 0x01, -1 } },
	{ desc_sprites },
	{ desc_END },
};



// ***********************
// Class 04:  CLIFF FACES
// ***********************

// (see simple vars above)


// ***********************
// Class 05:   WATER
// ***********************

varinf watervars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 1, 1, "costIndex" },
	{ 0x03, 1, 1, "var_03" },
	{ 0x04,-2, 1, "costFactor" },
	{ 0x06, 4, 1, "imageId" },
	{ 0x0A, 4, 1, "mapPixelImageId" },
	{ -1 },
};


// ***********************
// Class 06:  GROUND
// ***********************

varinf groundvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 1, 1, "costIndex" },
	{ 0x03, 1, 1, "var_03" },
	{ 0x04, 1, 1, "var_04" },
	{ 0x05, 1, 1, "flags" },
	{ 0x06, 1, 1, "cliffEdgeId1" },
	{ 0x07, 1, 1, "cliffEdgeId2" },
	{ 0x08,-2, 1, "costfactor" },
	{ 0x0A, 4, 1, "imageId" },
	{ 0x0E, 4, 1, "var_0E_ZERO" },
	{ 0x12, 4, 1, "cliffEdgeImageId" },
	{ 0x16, 4, 1, "mapPixelImageId" },
	{ 0x1A, 1, 1, "distributionPattern" },
	{ 0x1B, 1, 1, "numVariations" },
	{ 0x1C, 1, 1, "variationLikelihood" },
	{ 0x1D, 1, 1, "" },
	{ -1 },
};

objdesc grounddesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_useobj, { 0, ob_cliff, 0x04, -1 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 07:  TOWN NAMES
// ***********************

varinf townpart[] = {
	{ 0x00, 1, 1, "count" },
	{ 0x01, 1, 1, "bias" },
	{ 0x02, 2, 1, "offset" },
	{ -1 },
};

varinf townvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 4, 6, "categories", townpart },
	{ -1 },
};

objdesc towndesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_strtable, { 0, 2+0*4, 4+0*4 } },
	{ desc_strtable, { 1, 2+1*4, 4+1*4 } },
	{ desc_strtable, { 2, 2+2*4, 4+2*4 } },
	{ desc_strtable, { 3, 2+3*4, 4+3*4 } },
	{ desc_strtable, { 4, 2+4*4, 4+4*4 } },
	{ desc_strtable, { 5, 2+5*4, 4+5*4 } },
	{ desc_END },
};


// ***********************
// Class 08:  CARGO TYPES
// ***********************

const char *cargoflags[] = {
	"unk0", "refit", "delivering",
	NULL,
};

varinf cargovars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 1, "unitweight" },
	{ 0x04, 2, 1, "cargoTransferTime" },
	{ 0x06, 2, 1, "unitsAndCargoNameId" },
	{ 0x08, 2, 1, "unitNameSingularId" },
	{ 0x0A, 2, 1, "unitNamePluralId" },
	{ 0x0C, 4, 1, "unitInlineSpriteId" },
	{ 0x10, 2, 1, "cargoCategory" },
	{ 0x12, 1, 1, "flags", NULL, cargoflags },
	{ 0x13, 1, 1, "numPlatformVariations" },
	{ 0x14, 1, 1, "var_14" },
	{ 0x15, 1, 1, "premiumDays" },
	{ 0x16, 1, 1, "maxNonPremiumDays" },
	{ 0x17, 2, 1, "nonPremiumRate" },
	{ 0x19, 2, 1, "penaltyRate" },
	{ 0x1B, 2, 1, "paymentFactor" },
	{ 0x1D, 1, 1, "paymentIndex" },
	{ 0x1E, 1, 1, "unitSize" },
	{ -1 },
};

objdesc cargodesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_lang, { 1 } },
	{ desc_lang, { 2 } },
	{ desc_lang, { 3 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 09:    FENCES
// ***********************

const char *fenceflags[] = {
	"hasPrimaryColour", "hasGlass", "onlyOnLevelLand", "twoSided","unk4_NOT_USED","unk5_NOT_USED","hasSecondaryColour","hasTertirayColour_NOT_USED",
	NULL
};

varinf fencevars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 4, 1, "imageId" },
	{ 0x06, 1, 1, "cursorType_NOT_USED"},
	{ 0x07, 1, 1, "flags", NULL, fenceflags },
	{ 0x08, 1, 1, "height" },
	{ 0x09, 1, 1, "flags2_NOT_USED" },
	{ -1 },
};


// ***********************
// Class 0A:    SIGNALS
// ***********************

varinf signalvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 1, "flags"},
	{ 0x04, 1, 1, "animationSpeed"},
	{ 0x05, 1, 1, "numFrames"},
	{ 0x06,-2, 1, "costFactor" },
	{ 0x08,-2, 1, "sellCostFactor" },
	{ 0x0A, 1, 1, "costIndex" },
	{ 0x0B, 1, 1, "var_0B"},
	{ 0x0C, 2, 1, "descriptionId"},
	{ 0x0E, 4, 1, "imageId"},
	{ 0x12, 1, 1, "numCompatible"},
	{ 0x13, 1, 7, "modIds"},
	{ 0x1A, 2, 1, "designed" },
	{ 0x1C, 2, 1, "obsolete" },
	{ -1 },
};

objdesc signaldesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_lang, { 1 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 0B:  CROSSINGS
// ***********************

varinf crossingvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02,-2, 1, "costFactor" },
	{ 0x04,-2, 1, "sellCostFactor" },
	{ 0x06, 1, 1, "costIndex" },
	{ 0x07, 1, 1, "animationSpeed" },
	{ 0x08, 1, 1, "closingFrames" },
	{ 0x09, 1, 1, "closedFrames" },
	{ 0x0A, 1, 1, "var_0A" },
	{ 0x0B, 1, 1, "" },
	{ 0x0C, 2, 1, "designYear" },
	{ 0x0E, 4, 1, "imageId" },
	{ -1 },
};


// ***********************
// Class 0C: STREET LIGHTS
// ***********************

varinf lightvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 3, "designedYear" },
	{ 0x08, 4, 1, "imageId" },
	{ -1 },
};


// ***********************
// Class 0D:    TUNNELS
// ***********************

// (see simplevars above)


// ***********************
// Class 0E:   BRIDGES
// ***********************

varinf bridgevars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 1, 1, "flags" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 2, 1, "clearHeight" },
	{ 0x06, -2, 1, "deckDepth" },
	{ 0x08, 1, 1, "spanLength" },
	{ 0x09, 1, 1, "pillarSpacing" },
	{ 0x0A, 2, 1, "maxSpeed" },
	{ 0x0C, 1, 1, "maxHeight" },
	{ 0x0D, 1, 1, "costIndex" },
	{ 0x0E,-2, 1, "baseCostFactor" },
	{ 0x10,-2, 1, "heightCostFactor" },
	{ 0x12,-2, 1, "sellCostFactor" },
	{ 0x14, 2, 1, "disabledTrackCfg" },
	{ 0x16, 4, 1, "imageId" },
	{ 0x1A, 1, 1, "trackNumCompatible" },
	{ 0x1B, 1, 7, "trackModIds" },
	{ 0x22, 1, 1, "roadNumCompatible" },
	{ 0x23, 1, 7, "roadModIds" },
	{ 0x2A, 1, 2, "designedYear" },
	{ -1 },
};


// ***********************
// Class 0F:  TRAIN STATIONS
// ***********************

const char *track_pieces[] = {
	"diagonal", "widecurve", "mediumcurve", "smallcurve",	// 1 2 4 8
	"tightcurve", "normalslope", "steepslope", "oneSided",		// 10 20 40 80
	"slopedcurve", "sbend",	"junction",				// 100 200
	NULL,
};

varinf trnstatvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 1, 1, "drawStyle" },
	{ 0x03, 1, 1, "height" },
	{ 0x04, 2, 1, "trackpieces", NULL, track_pieces },
	{ 0x06,-2, 1, "buildCostFactor" },
	{ 0x08,-2, 1, "sellCostFactor" },
	{ 0x0A, 1, 1, "costIndex" },
	{ 0x0B, 1, 1, "var_0B" },
	{ 0x0C, 1, 1, "flags" },
	{ 0x0D, 1, 1, "var_0D" },
	{ 0x0E, 4, 1, "imageId" },
	{ 0x12, 4, 4, "var_12_Ids" },
	{ 0x22, 1, 1, "numCompatible" },
	{ 0x23, 1, 7, "modIds" },
	{ 0x2a, 2, 1, "designedYear" },
	{ 0x2c, 2, 1, "obsoleteYear" },
	{ 0x2E, 4, 4*4, "cargoOffsetBytes_ptrs" },
	{ 0x6E, 1, 16, "manualPower_ptrs" },
	{ -1 },
};

auxdesc trnstataux[1] = {
	{ "", NULL },
};

objdesc trnstatdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_auxdatavar, { 0, 32, 1, 4 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 10:  TRACK MODS
// ***********************

varinf trkmodvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 1, "trackPieces", NULL, track_pieces },
	{ 0x04, 1, 1, "paintSytle" },
	{ 0x05, 1, 1, "costIndex" },
	{ 0x06,-2, 1, "buildCostFactor" },
	{ 0x08,-2, 1, "sellCostFactor" },
	{ 0x0A, 4, 1, "imageId" },
	{ 0x0E, 4, 1, "var_0E_Id" },
	{ -1 },
};


// ***********************
// Class 11:   TRACKS
// ***********************

varinf trackvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 1, "trackPieces", NULL, track_pieces },
	{ 0x04, 2, 1, "stationTrackPieces", NULL, track_pieces },
	{ 0x06, 1, 1, "var_06" },
	{ 0x07, 1, 1, "numCompatible" },
	{ 0x08, 1, 1, "numMods" },
	{ 0x09, 1, 1, "numSignals" },
	{ 0x0A, 1, 4, "modIds" },
	{ 0x0E, 2, 1, "signalIds" },
	{ 0x10, 2, 1, "compatibleTracksIds" },
	{ 0x12, 2, 1, "compatibleRoadsIds" },
	{ 0x14,-2, 1, "buildCostFactor" },
	{ 0x16,-2, 1, "sellCostFactor" },
	{ 0x18,-2, 1, "tunnelCostFactor" },
	{ 0x1A, 1, 1, "costIndex" },
	{ 0x1B, 1, 1, "tunnelId" },
	{ 0x1C, 2, 1, "curveSpeed" },
	{ 0x1E, 4, 1, "imageId" },
	{ 0x22, 2, 1, "flags" },
	{ 0x24, 1, 1, "numBridges" },
	{ 0x25, 1, 7, "bridgeIds" },
	{ 0x2C, 1, 1, "numStations" },
	{ 0x2D, 1, 7, "stationIds" },
	{ 0x34, 1, 1, "displayOffset" },
	{ 0x35, 1, 1, "" },
	{ -1 },
};

objdesc trackdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_useobj, { -7, ob_compatible, 0x11, 0x14, -1 } },
	{ desc_useobj, { -8, ob_trackmod, 0x10, -1 } },
	{ desc_useobj, { -9, ob_signal, 0x0A, -1 } },
	{ desc_useobj, {  0, ob_tunnel, 0x0D, -1 } },
	{ desc_useobj, { -0x24, ob_bridge, 0x0E, -1 } },
	{ desc_useobj, { -0x2C, ob_station, 0x0F, -1 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 12:  ROAD STATIONS
// ***********************

const char *road_pieces[] = {
	"smallcurve", "tightcurve", "normalslope", "steepslope",
	"", "reverse",
	NULL,
};

varinf roadstvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 1, 1, "paintStyle" },
	{ 0x03, 1, 1, "height" },
	{ 0x04, 2, 1, "roadPieces", NULL, road_pieces },
	{ 0x06,-2, 1, "buildCostFactor" },
	{ 0x08,-2, 1, "sellCostFactor" },
	{ 0x0A, 1, 1, "costIndex" },
	{ 0x0B, 1, 1, "flags" },
	{ 0x0C, 4, 1, "imageId" },
	{ 0x10, 4, 4, "var_10_Ids" },
	{ 0x20, 1, 1, "numCompatible" },
	{ 0x21, 1, 7, "modIds" },
	{ 0x28, 2, 1, "designedYear" },
	{ 0x2a, 2, 1, "obsoleteYear" },
	{ 0x2C, 1, 1, "cargoTypeId" },
	{ 0x2D, 1, 1, "" },
	{ 0x2E, 4, 4*4, "cargoOffsetBytes_ptr" },
	{ -1 },
};

auxdesc roadstaux[1] = {
	{ "", NULL },
};

objdesc roadstdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_useobj, { 0, ob_cargo, 0x08, -1 } },
	{ desc_auxdatavar, { 0, 16, 1, 4 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 13:  ROAD MODS
// ***********************

// (same as track mods)


// ***********************
// Class 14:    ROADS
// ***********************

varinf roadvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 1, "roadPieces", NULL, road_pieces },
	{ 0x04,-2, 1, "buildCostFactor" },
	{ 0x06,-2, 1, "sellCostFactor" },
	{ 0x08,-2, 1, "tunnelCostFactor" },
	{ 0x0A, 1, 1, "costIndex" },
	{ 0x0B, 1, 1, "tunnelId" },
	{ 0x0C, 2, 1, "maxSpeed" },
	{ 0x0E, 4, 1, "imageId" },
	{ 0x12, 2, 1, "flags" },
	{ 0x14, 1, 1, "numBridges" },
	{ 0x15, 1, 7, "bridgeIds" },
	{ 0x1C, 1, 1, "numStations" },
	{ 0x1D, 1, 7, "stationIds" },
	{ 0x24, 1, 1, "paintStyle" },
	{ 0x25, 1, 1, "numMods" },
	{ 0x26, 1, 2, "modIds" },
	{ 0x28, 1, 1, "numCompatible" },
	{ 0x29, 1, 1, "displayOffset" },
	{ 0x2A, 2, 1, "compatibleRoadsIds" },
	{ 0x2C, 2, 1, "compatibleTracksIds" },
	{ 0x2E, 1, 1, "targetTownSize" },
	{ 0x2F, 1, 1, "" },
	{ -1 },
};

objdesc roaddesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_useobj, { -40, ob_compatible, 0x11, 0x14, -1 } },
	{ desc_useobj, { -37, ob_roadmod, 0x13, -1 } },
	{ desc_useobj, {  0, ob_tunnel, 0x0D, -1 } },
	{ desc_useobj, { -20, ob_bridge, 0x0E, -1 } },
	{ desc_useobj, { -28, ob_station, 0x12, -1 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 15:   AIRPORTS
// ***********************

varinf common_building_part_height[] = {
	{ 0x00, 1, 1, "height" },
	{ -1 },
};

varinf animationvars[] = {
	{ 0x00, 1, 1, "numFrames" },
	{ 0x01, 1, 1, "animationSpeed" },
	{ -1 },
};

varinf common_building_part_animation[] = {
	{ 0x00, 2, 0, "animation", animationvars },
	{ -1 },
};

varinf common_building_variation_parts[] = {
	{ 0x00, 1, 0, "part" },
	{ -1 },
};

varinf layoutvars[] = {
	{ 0x00, 1, 1, "index" },
	{ 0x01, 1, 1, "rotation" },
	{ 0x02,-1, 1, "x" },
	{ 0x03,-1, 1, "y" },
	{ -1 },
};

varinf airport_aux3[] = {
	{ 0x00, 4, 0, "tilepos", layoutvars },
	{ -1 },
};

const char *airportaux4flags[] = {
	"terminal", "aircraftend", "", "ground",
	"flight", "helibegin", "aircraftbegin", "heliend",
	"touchdown",
	NULL,
};

varinf airport_aux4[] = {
	{ 0x00,-2, 1, "x" },
	{ 0x02,-2, 1, "y" },
	{ 0x04,-2, 1, "z" },
	{ 0x06, 2, 1, "flags", NULL, airportaux4flags },
	{ -1 },
};

varinf airport_aux5[] = {
	{ 0x00, 1, 1, "var_00" },
	{ 0x01, 1, 1, "curNode" },
	{ 0x02, 1, 1, "nextNode" },
	{ 0x03, 1, 1, "var_03" },
	{ 0x04, 4, 1, "mustBeClearEdges", NULL, nobits },
	{ 0x08, 4, 1, "atLeastOneClearEdges", NULL, nobits },
	{ -1 },
};

varinf airportvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02,-2, 1, "buildCostFactor" },
	{ 0x04,-2, 1, "sellCostFactor" },
	{ 0x06, 1, 1, "costIndex" },
	{ 0x07, 1, 1, "var_07" },
	{ 0x08, 4, 1, "imageId" },
	{ 0x0C, 4, 1, "buildingImageId" },
	{ 0x10, 2, 1, "flags" },
	{ 0x12, 1, 1, "numSpriteSets" },
	{ 0x13, 1, 1, "numTiles" },
	{ 0x14, 4, 1, "buildingPartHeights_ptr" },
	{ 0x18, 4, 1, "buildingPartAnimations_ptr" },
	{ 0x1C, 4, 32, "buildingVariationParts_ptrs" },
	{ 0x9C, 4, 1, "buildingPositions_ptr" },
	{ 0xA0, 4, 1, "largeTiles", NULL, nobits  },
	{ 0xA4,-1, 1, "minX" },
	{ 0xA5,-1, 1, "minY" },
	{ 0xA6,-1, 1, "maxX" },
	{ 0xA7,-1, 1, "maxY" },
	{ 0xA8, 2, 1, "designedYear" },
	{ 0xAA, 2, 1, "obsoleteYear" },
	{ 0xAC, 1, 1, "numMovementNodes" },
	{ 0xAD, 1, 1, "numMovementEdges" },
	{ 0xAE, 4, 1, "movementNodes_ptr" },
	{ 0xB2, 4, 1, "movementEdges_ptr" },
	{ 0xB6, 4, 1, "var_B6", NULL, nobits },
	{ -1 },
};

auxdesc airportaux[6] = {
	{ "buildingPartHeights", common_building_part_height },
	{ "buildingPartAnimations", common_building_part_animation },
	{ "buidlingVarationParts", common_building_variation_parts },
	{ "buildingPositions", airport_aux3 },
	{ "movementNodes", airport_aux4 },
	{ "movementEdges", airport_aux5},
};

objdesc airportdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_auxdata, { 0, 0, 1, -0x12 } },
	{ desc_auxdata, { 1, 0, -2, -0x12 } },
	{ desc_auxdatavar, { 2, -0x13, 1, 1 } },
	{ desc_auxdatavar, { 3, 0, -4, 1 } },
	{ desc_auxdata, { 4, -0xac, 1, 8 } },
	{ desc_auxdata, { 5, -0xad, 1, 12 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 16:     DOCKS
// ***********************

varinf dockvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02,-2, 1, "buildCostFactor" },
	{ 0x04,-2, 1, "sellCostFactor" },
	{ 0x06, 1, 1, "costIndex" },
	{ 0x07, 1, 1, "var_07" },
	{ 0x08, 4, 1, "imageId" },
	{ 0x0C, 4, 1, "buildingImageId" },
	{ 0x10, 2, 1, "flags" },
	{ 0x12, 1, 1, "numBuildingParts" },
	{ 0x13, 1, 1, "numBuildingVariations" },
	{ 0x14, 4, 1, "buildingPartHeights_ptr" },
	{ 0x18, 4, 1, "buildingPartAnimations_ptr" },
	{ 0x1C, 4, 1, "buildingVariationParts_ptrs" },
	{ 0x20, 2, 1, "designedYear" },
	{ 0x22, 2, 1, "obsoleteYear" },
	{ 0x24, 2, 1, "boatPosition_x" },
	{ 0x26, 2, 1, "boatPosition_y" },
	{ -1 },
};

auxdesc dockaux[3] = {
	{ "buildingPartHeights", common_building_part_height },
	{ "buildingPartAnimations", common_building_part_animation },
	{ "buidlingVarationParts", common_building_variation_parts },
};

objdesc dockdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_auxdata, { 0, 0, 1, -0x12 } },
	{ desc_auxdata, { 1, 0, -2, -0x12 } },
	{ desc_auxdatavar, { 2, -0x13, 1, 1 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 17:   VEHICLES
// ***********************


	// the known bits in the vehicle sprite structure flags
const char *vehspriteflags[] = {
	"hasSprites", "rotationalSymmetry", "hasUnkSprites", "hasGentleSprites", "hasSteepSprites",
	"hasBrakingLights", "hasSpeedAnimation",
	NULL
};

	// the vehicle sprite definition structure
varinf vehsprites[] = {
	{ 0x00, 1, 1, "numFlatRotationFrames" },
	{ 0x01, 1, 1, "numSlopedRotationFrames" },
	{ 0x02, 1, 1, "numAnimationFrames" },
	{ 0x03, 1, 1, "numCargoLoadFrames" },
	{ 0x04, 1, 1, "numCargoFrames" },
	{ 0x05, 1, 1, "numRollFrames" },
	{ 0x06, 1, 1, "halfLength" },
	{ 0x07, 1, 1, "flags", NULL, vehspriteflags },
	{ 0x08, 1, 1, "width" },
	{ 0x09, 1, 1, "heightNegative" },
	{ 0x0A, 1, 1, "heightPositive" },
	{ 0x0B, 1, 1, "flatYawAccuracy" },
	{ 0x0C, 1, 1, "slopedYawAccuracy" },
	{ 0x0D, 1, 1, "numFramesPerRotation" },
	{ 0x0E, 4, 1, "flatImageId" },
	{ 0x12, 4, 1, "unkImageId" },
	{ 0x16, 4, 1, "gentleImageId" },
	{ 0x1A, 4, 1, "steepImageId" },
	{ -1 },
};
varinf vehbogiesprites[] = {
	{ 0x00, 1, 1, "numAnimationFrames" },
	{ 0x01, 1, 1, "flags", NULL, vehspriteflags },
	{ 0x02, 1, 1, "width" },
	{ 0x03, 1, 1, "heightNegative" },
	{ 0x04, 1, 1, "heightPositive" },
	{ 0x05, 1, 1, "numFramesPerRotation" },
	{ 0x06, 4, 1, "flatImageId" },
	{ 0x0A, 4, 1, "gentleImageId" },
	{ 0x0E, 4, 1, "steepImageId" },
	{ -1 },
};

	// a vehicle structure whose purpose is mostly unknown
varinf vehunk[] = {
	{ 0x00, 1, 1, "frontBogiePosition" },
	{ 0x01, 1, 1, "backBogiePosition" },
	{ 0x02, 1, 1, "frontBogieSpriteInd" },
	{ 0x03, 1, 1, "backBogieSpriteInd" },
	{ 0x04, 1, 1, "bodySpriteInd" },
	{ 0x05, 1, 1, "var_05" },
	{ -1 },
};

	// the vehicle flags
const char *vehflags[] = {
	"", "", "", "",
	"", "", "rackrail", "",
	"", "anytrack", "", "cancouple",
	"dualhead", "", "refittable", "noannounce",
	NULL
};

	// the main vehicle data structure
varinf vehvars[] = {
	{ 0x000, 2, 1, "nameId" },
	{ 0x002, 1, 1, "mode" },
	{ 0x003, 1, 1, "type" },
	{ 0x004, 1, 1, "var_04" },
	{ 0x005, 1, 1, "trackTypeId" },
	{ 0x006, 1, 1, "numMods" },
	{ 0x007, 1, 1, "costIndex" },
	{ 0x008,-2, 1, "costFactor" },	// size -2 means signed word
	{ 0x00A, 1, 1, "reliability" },
	{ 0x00B, 1, 1, "runCostIndex" },
	{ 0x00C,-2, 1, "runCostFactor" },
	{ 0x00E, 1, 1, "colourType" },
	{ 0x00F, 1, 1, "numCompatibleVehicles" },
	{ 0x010, 2, 8, "compatibleVehicleIds" },
	{ 0x020, 1, 4, "requiredTrackExtras" },
	{ 0x024, 6, 4, "carComponents", vehunk },
	{ 0x03C,30, 4, "bodySprites", vehsprites },
	{ 0x0B4,18, 2, "bogieSprites", vehbogiesprites },
	{ 0x0D8, 2, 1, "power" },
	{ 0x0DA, 2, 1, "speed" },
	{ 0x0DC, 2, 1, "rackSpeed" },
	{ 0x0DE, 2, 1, "weight" },
	{ 0x0E0, 2, 1, "flags", NULL, vehflags },
	{ 0x0E2, 1, 2, "maxCargo" },
	{ 0x0E4, 4, 2, "compatibleCargoCategories" },
	{ 0x0EC, 1,32, "cargoTypeSpriteOffsets" },
	{ 0x10C, 1, 1, "numSimultaneousCargoTypes" },
	{ 0x10D, 3, 2, "animations" },
	{ 0x113, 1, 1, "var_113" },
	{ 0x114, 2, 1, "designed" },
	{ 0x116, 2, 1, "obsolete" },
	{ 0x118, 1, 1, "rackRailTypeId" },
	{ 0x119, 1, 1, "drivingSoundType" },
	{ 0x11A, 27,1, "sound" },
	{ 0x135, 1,37, "" },
	{ 0x15A, 1, 1, "numStartSounds" },
	{ 0x15B, 1, 3, "startSoundIds" },
	{ -1 },
};

objdesc vehdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_useobj, { descnumspec(0), ob_tracktype, 0x11, 0x14, -1 } },
	{ desc_useobj, { -6, ob_trackmod, 0x10, 0x13, -1 } },
	{ desc_cargo, { 2 } },
	{ desc_useobj, { descnumif(0x10f), ob_visualeffect, 0x03, -1 } },
	{ desc_useobj, { descnumif(0x112), ob_wakeeffect, 0x03, -1 } },
	{ desc_useobj, { descnumspec(1), ob_rackrail, 0x10, -1 } },
	{ desc_useobj, { -15, ob_compatible, 0x17, -1 } },
	{ desc_useobj, { descnumif(0x119), ob_startsnd, 0x01, -1 } },
	{ desc_useobj, { descnumand(0x15a,0x7f), ob_soundeffect, 0x01, -1 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 18:    TREES
// ***********************

varinf treevars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 1, 1, "initialHeight" },
	{ 0x03, 1, 1, "height" },
	{ 0x04, 1, 1, "var_04" },
	{ 0x05, 1, 1, "var_05" },
	{ 0x06, 1, 1, "numRotations" },
	{ 0x07, 1, 1, "growth" },
	{ 0x08, 2, 1, "flags" },
	{ 0x0A, 4, 6, "spriteIds" },
	{ 0x22, 4, 6, "snowSpriteIds" },
	{ 0x3A, 2, 1, "shadowImageOffset" },
	{ 0x3C, 1, 1, "var_3C" },
	{ 0x3D, 1, 1, "seasonState" },
	{ 0x3E, 1, 1, "var_3E" },
	{ 0x3F, 1, 1, "costIndex" },
	{ 0x40,-2, 1, "buildCostFactor" },
	{ 0x42,-2, 1, "clearCostFactor" },
	{ 0x44, 4, 1, "colours" },
	{ 0x48,-2, 1, "rating" },
	{ 0x4A,-2, 1, "demolishRatingReduction" },
	{ -1 },
};


// ***********************
// Class 19:    SNOW
// ***********************

// (see simple vars above)


// ***********************
// Class 1A:   CLIMATES
// ***********************

varinf climvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 1, 1, "firstSeason" },
	{ 0x03, 1, 4, "seasonLength" },
	{ 0x07, 1, 1, "winterSnowLine" },
	{ 0x08, 1, 1, "summerSnowLine" },
	{ 0x09, 1, 1, "" },
	{ -1 },
};

objdesc climdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_END },
};


// ***********************
// Class 1B:   SHAPES
// ***********************

varinf shapevars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 1, 1, "hillHeightMapCount" },
	{ 0x03, 1, 1, "mountainHeightMapCount" },
	{ 0x04, 4, 1, "imageId" },
	{ 0x08, 4, 1, "imageHillsId" },
	{ 0x0C, 2, 1, "flags" },
	{ -1 },
};


// ***********************
// Class 1C:   BUILDINGS
// ***********************

const char *bldg_flags[] = {
	"", "", "", "ishq",	// 1 2 4 8
	NULL,
};

varinf bldngvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 4, 1, "imageId" },
	{ 0x06, 1, 1, "numParts" },
	{ 0x07, 1, 1, "numVariations" },
	{ 0x08, 4, 1, "buildingPartHeights_ptr" },
	{ 0x0C, 4, 1, "buildingPartAnimations_ptr" },
	{ 0x10, 4, 32, "buildingVariationParts_ptrs" },
	{ 0x90, 4, 1, "colours" },
	{ 0x94, 2, 1, "designedYear" },
	{ 0x96, 2, 1, "obsoleteYear" },
	{ 0x98, 1, 1, "flags", NULL, bldg_flags },
	{ 0x99, 1, 1, "clearCostIndex" },
	{ 0x9A,-2, 1, "clearCostFactor" },
	{ 0x9C, 1, 1, "scaffoldingSegmentTypeId" },
	{ 0x9D, 1, 1, "scaffoldingColour" },
	{ 0x9E, 1, 1, "generatorFunction" },
	{ 0x9F, 1, 1, "averageNumberOnMap" },
	{ 0xA0, 1, 2, "producedQuantity" },
	{ 0xA2, 1, 2, "producedCargoType" },
	{ 0xA4, 1, 2, "requiredCargoType" },
	{ 0xA6, 1, 2, "var_A6" },
	{ 0xA8, 1, 2, "var_A8" },
	{ 0xAA,-2, 1, "demolishRatingReduction" },
	{ 0xAC, 1, 1, "var_AC" },
	{ 0xAD, 1, 1, "numElevatorSequences" },
	{ 0xAE, 4, 4, "elevatorHeightSequence_ptrs" },
	{ -1 },
};

auxdesc bldngaux[4] = {
	{ "buildingPartHeights", common_building_part_height },
	{ "buildingPartAnimations", common_building_part_animation },
	{ "buidlingVarationParts", common_building_variation_parts },
	{ "elevatorHeightSequences", NULL },
};

objdesc bldngdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_auxdata, { 0, 0, 1, -6 } },
	{ desc_auxdata, { 1, 0, -2, -6 } },
	{ desc_auxdatavar, { 2, -7, 1, 1 } },
	{ desc_useobj, { 4, ob_cargo, 0x08, 0xFF, -1 } },
	{ desc_auxdatafix, { 3, -0xad, 1, 2 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 1D:  SCAFFOLDING
// ***********************

varinf scaffvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 4, 1, "imageId" },
	{ 0x06, 2, 3, "segmentHeights" },
	{ 0x0C, 2, 3, "roofHeights" },
	{ -1 },
};


// ***********************
// Class 1E:  INDUSTRIES
// ***********************

const char *industryflags[] = {
	"", "", "", "",		// 1, 2, 4, 8
	"", "", "", "",		// 10, 20, 40, 80
	"", "", "", "",		// 100, 200, 400, 800
	"", "", "", "",		// 1000, 2000, 4000, 8000
	"", "needall", "canincreaseproduction", "candecreaseproduction",
				// 100000, 200000, 400000, 800000
	NULL,
};

varinf indproduction_rate_range[] = {
	{ 0x00, 2, 1, "min" },
	{ 0x02, 2, 1, "max" },
	{ -1 },
};

varinf indvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 1, "var_02Id" },
	{ 0x04, 2, 1, "nameClosingDownId" },
	{ 0x06, 2, 1, "nameUpProductionId" },
	{ 0x08, 2, 1, "nameDownProductionId" },
	{ 0x0A, 2, 1, "nameSingularId" },
	{ 0x0C, 2, 1, "namePluralId" },
	{ 0x0E, 4, 1, "shadowImageId" },
	{ 0x12, 4, 1, "baseImageId" },
	{ 0x16, 4, 1, "var_16_Id" },
	{ 0x1A, 4, 1, "var_1A_Id" },
	{ 0x1E, 1, 1, "numBuildingParts" },
	{ 0x1F, 1, 1, "numBuildingVariations" },
	{ 0x20, 4, 1, "buildingPartHeights_ptr" },
	{ 0x24, 4, 1, "buildingPartAnimations_ptr" },
	{ 0x28, 4, 4, "animationSequence_ptrs" },
	{ 0x38, 4, 1, "var_38_ptrs" },
	{ 0x3C, 4, 32, "buildingVariationPart_ptrs" },
	{ 0xBC, 1, 1, "minNumBuildings" },
	{ 0xBD, 1, 1, "maxNumBuildings" },
	{ 0xBE, 4, 1, "buildings_ptr" },
	{ 0xC2, 4, 1, "availableColours" },
	{ 0xC6, 4, 1, "buildingSizeFlags" },
	{ 0xCA, 2, 1, "designedYear" },
	{ 0xCC, 2, 1, "obsoleteYear" },
	{ 0xCE, 1, 1, "totalOfTypeInScenario" },
	{ 0xCF, 1, 1, "costIndex" },
	{ 0xD0,-2, 1, "costFactor" },
	{ 0xD2,-2, 1, "clearCostFactor" },
	{ 0xD4, 1, 1, "scaffoldingSegmentTypeId" },
	{ 0xD5, 1, 1, "scaffoldingColour" },
	{ 0xD6, 4, 2, "initialProductionRate", indproduction_rate_range },
	{ 0xDE, 1, 2, "producedCargoTypeIds" },
	{ 0xE0, 1, 3, "requiredCargoTypeIds" },
	{ 0xE3, 1, 1, "mapColour" },
	{ 0xE4, 4, 1, "flags", NULL, industryflags },
	{ 0xE8, 1, 1, "var_E8" },
	{ 0xE9, 1, 1, "var_E9" },
	{ 0xEA, 1, 1, "var_EA" },
	{ 0xEB, 1, 1, "var_EB" },
	{ 0xEC, 1, 1, "var_EC" },
	{ 0xED, 1, 4, "wallTypeIds" },
	{ 0xF1, 1, 1, "buildingWallId" },
	{ 0xF2, 1, 1, "buildingWallEntranceId" },
	{ 0xF3, 1, 1, "var_F3" },
	{ -1 },
};

auxdesc indaux[6] = {
	{ "buildingPartHeights", common_building_part_height },
	{ "buildingPartAnimations", common_building_part_animation },
	{ "animationSequences", NULL },
	{ "var_38s", NULL },
	{ "buidlingVarationParts", common_building_variation_parts },
	{ "buildings", NULL },
};

objdesc inddesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_lang, { 1 } },
	{ desc_lang, { 2 } },
	{ desc_lang, { 3 } },
	{ desc_lang, { 4 } },
	{ desc_lang, { 5 } },
	{ desc_lang, { 6 } },
	{ desc_lang, { 7 } },
	{ desc_auxdata, { 0, 0, 1, -30 } },
	{ desc_auxdata, { 1, 0, 2, -30 } },
	{ desc_auxdatafix, { 2, 4, 1, 1 } },
	{ desc_auxdatavar, { 3, 0, 2, 1 } },
	{ desc_auxdatavar, { 4, -31, 1, 1 } },
	{ desc_auxdata, { 5, 0, 1, -0xbd } },
	{ desc_useobj, { 2, ob_produces, 0x08, 0xFF, -1 } },
	{ desc_useobj, { 3, ob_accepts, 0x08, 0xFF, -1 } },
	{ desc_useobj, { 6, ob_fence, 0x09, 0xFF, -1 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 1F:   REGIONS
// ***********************

varinf regionvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 4, 1, "imageId" },
	{ 0x06, 1, 2, "" },
	{ 0x08, 1, 1, "numCargoInflunceObjects" },
	{ 0x09, 1, 4, "cargoInfluenceObjectIds" },
	{ 0x0D, 1, 4, "cargoInfluenceTownFilter" },
	{ 0x11, 1, 1, "" },
	{ -1 },
};

objdesc regiondesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_useobj, { -8, ob_cargo, 0x08, -1 } },
	{ desc_useobj, { descnumuntil(), ob_default, 0x100, -1 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 20:  COMPANIES
// ***********************

const char *comp_sprites[] = {
	NULL,
};

varinf compvars[] = {
	{ 0x00, 2, 1, "nameId" },
	{ 0x02, 2, 1, "lastNameId" },
	{ 0x04, 4, 1, "var_04" },
	{ 0x08, 4, 1, "var_08" },
	{ 0x0C, 4, 1, "emotions", NULL, comp_sprites },
	{ 0x10, 4, 9, "imageIds" },
	{ 0x34, 1, 1, "intelligence" },
	{ 0x35, 1, 1, "aggressiveness" },
	{ 0x36, 1, 1, "competitiveness" },
	{ 0x37, 1, 1, "var_37" },
	{ -1 },
};

objdesc compdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_lang, { 1 } },
	{ desc_sprites },
	{ desc_END },
};


// ***********************
// Class 21:   TEXTS
// ***********************

// (uses simple vars)

objdesc textdesc[] = {
	{ desc_objdata },
	{ desc_lang, { 0 } },
	{ desc_lang, { 1 } },
	{ desc_END },
};


// ***********************
//  END OF CLASS HANDLERS
// ***********************

//
// Array defining object class-specific data structures and descriptions
//
objclass objclasses[NUMOBJCLASSES] = {
	// structure	size auxdef	description
	{ interfacevars, 24, NULL,	simpledesc },	// 00 Interfaces
	{ sfxvars,	 12, NULL,	sfxdesc },	// 01 Sound effects
	{ currvars,	 12, NULL,	currdesc },	// 02 Currencies
	{ exhfxvars,	 40, exhfxaux,	exhfxdesc },	// 03 Exhaust effects
	{ simplevars,	  6, NULL,	simpledesc },	// 04 Cliff faces
	{ watervars,	 14, NULL,	simpledesc },	// 05 Water
	{ groundvars,	 30, NULL,	grounddesc },	// 06 Ground
	{ townvars,	 26, NULL,	towndesc },	// 07 Town names
	{ cargovars,	 31, NULL,	cargodesc },	// 08 Cargos
	{ fencevars,	 10, NULL,	simpledesc },	// 09 Fences
	{ signalvars,	 30, NULL,	signaldesc },	// 0A Signals
	{ crossingvars,	 18, NULL,	simpledesc },	// 0B Crossings
	{ lightvars,	 12, NULL,	simpledesc },	// 0C Street lights
	{ simplevars,	  6, NULL,	simpledesc },	// 0D Tunnels
	{ bridgevars,	 44, NULL,	simpledesc },	// 0E Bridges
	{ trnstatvars,	174, trnstataux,trnstatdesc },	// 0F Train stations
	{ trkmodvars,	 18, NULL,	simpledesc },	// 10 Track modifications
	{ trackvars,	 54, NULL,	trackdesc },	// 11 Tracks
	{ roadstvars,	110, roadstaux,	roadstdesc },	// 12 Road stations
	{ trkmodvars,	 18, NULL,	simpledesc },	// 13 Road modifications
	{ roadvars,	 48, NULL,	roaddesc },	// 14 Roads
	{ airportvars,	186, airportaux,airportdesc },	// 15 Airports
	{ dockvars,	 40, dockaux,	dockdesc },	// 16 Docks
	{ vehvars,	350, NULL,	vehdesc },	// 17 Vehicles
	{ treevars,	 76, NULL,	simpledesc },	// 18 Trees
	{ simplevars,	  6, NULL,	simpledesc },	// 19 Snow
	{ climvars,	 10, NULL,	climdesc },	// 1A Climates
	{ shapevars,	 14, NULL,	simpledesc },	// 1B Shapes
	{ bldngvars,	190, bldngaux,	bldngdesc },	// 1C bldngs
	{ scaffvars,	 18, NULL,	simpledesc },	// 1D Scaffolding
	{ indvars,	244, indaux,	inddesc },	// 1E Industries
	{ regionvars,	 18, NULL,	regiondesc },	// 1F Regions
	{ compvars,	 56, NULL,	compdesc },	// 20 Companies
	{ simplevars,	  6, NULL,	textdesc },	// 21 Texts
};
