//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// options.c - code for option settings
//

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

#include "common.h"
#include "locodef.h"

#ifdef __linux
#	define __declspec(a)
#	define mkdir(a) mkdir(a, 0777)
#endif
#ifdef __windows
#	define WIN32_LEAN_AND_MEAN
#	include <windows.h>
#endif

// Global variables

int embedpng = 0;	// can't be set atm, only 0 supported
int allinone = 0;	// put all images in the same png file
int demofile = 0;	// if 1, generate checksum for the demo
int palettetype = 27;	// which colour palette to use for company colours
int quiet = 0;		// no output, no messages, fail silently
int heapdebug = 0;	// to debug the heap functions
int allunknown = 0;	// if 1, print all unknown variables, even empty ones
int onlysetbits;	// if 1, don't generate unset bits, even known ones

char *_filebase = NULL;	// filename with the extension removed
char *_pngbase = NULL;	// same, but as directory

#ifdef __windows
__declspec(dllexport) BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fwdReason, LPVOID lpvReserved)
{
/*
  FILE *log = fopen("locotool.dll.log", "at");
  fprintf(log, "locotool.dll DllMain(%p, %lx, %p) called\n", hinstDLL, fwdReason, lpvReserved);
  fclose(log);
*/
  return TRUE;
}
#endif

struct {
	char **s;
	int set;
} stroptions[opt_STRLAST] = {
	{ &_filebase },
	{ &_pngbase },
};

struct {
	int *v, set;
} intoptions[opt_INTLAST] = {
	{ &palettetype },
	{ &demofile },
	{ &quiet },
	{ &allinone },
	{ &heapdebug },
	{ &allunknown },
	{ &onlysetbits },

};

char *LocoTool_set_str_option(en_stroption opt, char *str)
{
  char *old;
  if (opt >= opt_STRLAST)
	die("String option %d out of range", opt);

  if (stroptions[opt].set == 2)
	die("Too late to set string option %d", opt);

  old = *stroptions[opt].s;
  *stroptions[opt].s = str;
  stroptions[opt].set = 1;

  return old;
}

int LocoTool_set_str_option_copy(en_stroption opt, char *str)
{
  char *old;
  if (opt >= opt_STRLAST)
	die("String option %d out of range", opt);

  if (stroptions[opt].set == 2)
	die("Too late to set string option %d", opt);

  if (stroptions[opt].set)
	my_free(*stroptions[opt].s);

  if (str) {
	*stroptions[opt].s = my_strdup(str, "str_option");
	stroptions[opt].set = 1;
  } else {
	stroptions[opt].set = 0;
  }

  return 1;
}

char *LocoTool_get_str_option(en_stroption opt)
{
  if (opt >= opt_STRLAST)
	die("String option %d out of range", opt);

  return *stroptions[opt].s;
}

int LocoTool_set_int_option(en_stroption opt, int val)
{
  int old;

  if (opt >= opt_INTLAST)
	die("Integer option %d out of range", opt);

  old = *intoptions[opt].v;
  *intoptions[opt].v = val;
  intoptions[opt].set = 1;

  return old;
}

int LocoTool_get_int_option(en_stroption opt)
{
  if (opt >= opt_INTLAST)
	die("Integer option %d out of range", opt);

  return *intoptions[opt].v;
}

void clearoptions(void)
{
  if (stroptions[opt_filebase].set >= 2) {
	my_free(_filebase);
	stroptions[opt_filebase].set = 0;
  }
  if (stroptions[opt_pngbase].set >= 2) {
	my_free(_pngbase);
	stroptions[opt_pngbase].set = 0;
  }
}

char *makefilename(const char *suffix)
{
  char *name = my_alloc(strlen(*stroptions[opt_filebase].s)+strlen(suffix)+1, suffix);

  strcpy(name, *stroptions[opt_filebase].s);
  strcat(name, suffix);

  return name;
}

char *makepngname(int num)
{
  struct stat st;
  char *name;

  if (stroptions[opt_pngbase].set == 2) {
	if (stat(_pngbase, &st))
		if (mkdir(_pngbase))
			die("Cannot make directory %s: %s", _pngbase, strerror(errno));
#ifdef __linux
	  strcat(_pngbase, "/");
#else
	  strcat(_pngbase, "\\");
#endif

	stroptions[opt_pngbase].set = 3;
  }

  name = my_alloc(strlen(*stroptions[opt_pngbase].s)+16, "pngname");

  strcpy(name, *stroptions[opt_pngbase].s);
  sprintf(name+strlen(name), "%03d.png", num);

  return name;
}

// store the filename with the extension removed
void setfilebase(char *fname)
{
  if (!stroptions[opt_filebase].set) {
	_filebase = my_alloc(strlen(fname)+1, "filebase");

	strcpy(_filebase, fname);
	if (strrchr(_filebase, '.'))
		strrchr(_filebase, '.')[0] = 0;

	stroptions[opt_filebase].set = 2;
  }

  if (!stroptions[opt_pngbase].set) {
	_pngbase = my_alloc(strlen(_filebase)+2, "pngbase");

	strcpy(_pngbase, _filebase);

	stroptions[opt_pngbase].set = 2;
  }
}
