//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// png.c - functions to read and write png files using libpng
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <png.h>

#include "common.h"
#include "locodef.h"

int bgcol = 255;	// transparent ("chunked") background
int textcol = 10;	// text colour for sprite numbers
int framecol = 254;	// colour for frame around sprites
png_byte trnscol = 0;

#ifdef DEBUGDMP
int colhist[256] = { -1 };
#endif

//
// Take a row of sprite data and assemble it into a PNG pixel row
//
// For sprites that are chunked, the "holes" in the data are transparent,
// and we use colour bgcol to indicate that.
//
// Sprites that aren't chunked are simply copied over.
//
uint32_t putspriterow(uint8_t *data, int flags, int y, int width, uint8_t *row)
{
  uint32_t offset, last, ofs, len;

#ifdef DEBUGDMP
  uint32_t x1;

  if (colhist[0] == -1)
	memset(colhist, 0, sizeof(colhist));
#endif

  memset(row, bgcol, width);

  if (!(flags&4)) {
	memcpy(row, data+y*width, width);

#ifdef DEBUGDMP
	for (x1=0; x1<width; x1++)
		colhist[row[x1]]++;
#endif

	return (y+1)*width;
  }

  offset=getvalue(data+2*y, 2);

  do {
	last=data[offset]  & 0x80;
	len=data[offset++] & 0x7f;
	ofs=data[offset++];

	memcpy(row+ofs, data+offset, len);

#ifdef DEBUGDMP
	for (x1=ofs; x1<ofs+len; x1++)
		colhist[row[x1]]++;
#endif

	offset+=len;
  } while (!last);

  return offset;
}


//
// Reverse of the above, take a row of PNG pixels and encode it as sprite data
//
// For chunked sprites, we also assemble the data offset array
//
uint32_t getspriterow(uint8_t *data, uint32_t size, int flags, int y, int width, uint8_t *row)
{
  uint32_t offset, x1, x2, len;

  if (!(flags&4)) {
	memcpy(data+y*width, row, width);
	return (y+1)*width;
  }

  ((uint16_t*) data)[y] = size;

  x1=x2=0;

  offset = size;
  while (x1 < width) {
	while ( (x1 < width) && (row[x1]==bgcol))
		x1++;

	if (x1 < width) {
		len = 1;
		x2 = x1+len;

		while ((x2 < width) && (len < 127) && (row[x2]!=bgcol)) {
			x2++;
			len++;
		}

		offset = size;
		data[size++] = len;
		data[size++] = x1;

		memcpy(data+size, row+x1, len);
		size += len;

	} else {	// transparent to end of line
		if (x2 == 0) {	// totally empty line
			data[size++] = 0;
			data[size++] = 0;
		}
		x2 = x1;
	}
	x1 = x2;
  }

  data[offset] |= 0x80;

  return size;
}

//
// IO functions called by the PNG writer
//
void xml_write(png_structp png_ptr, png_bytep data, png_size_t size)
{
  fwrite(data, 1, size, png_get_io_ptr(png_ptr));
}

void xml_flush(png_structp png_ptr)
{
  fflush(png_get_io_ptr(png_ptr));
}

#include "locopal.h"

// for all-in-one png files the pixel rows are stored here
int allocrows = 0;
int numrows = 0;
int thisx;
int nextx = 65535;
int lastnumx = -200;
int thisy;
int nexty = 8;
png_byte **rows;

#include "digitbmp.h"

uint32_t addsprite(uint8_t *data, int id, uint32_t width, uint32_t height, uint32_t flags, int *x, int *y)
{
  int i, j;
  char spritenum[32];
  uint32_t dumped;

  dumped = 0;

  thisx = nextx;
  nextx = (thisx + width + 7) & ~4;
  if (nextx > 800) {
	thisx = 4;
	thisy = nexty;
	nextx = (thisx + width + 7) & ~4;
	lastnumx = -200;
  }

  i = (thisy + height + 11) & ~4;
  if (i > nexty)
	nexty = i;

  for (i=numrows; i<nexty; i++) {
	if (i >= allocrows) {
		j = allocrows;
		allocrows += 256;

		if (j)
			rows = my_realloc(rows, allocrows*sizeof(rows[0]), j*sizeof(rows[0]), "rows");
		else
			rows = my_alloc(allocrows*sizeof(rows[0]), "rows");
	}

	rows[i] = my_alloc(800, "row");
	memset(rows[i], framecol, 800);
	numrows = i+1;
  }

  // add the sprite number
  snprintf(spritenum, sizeof(spritenum)-1, "%d", id);
  if (thisx > lastnumx) {
	for (i=0; spritenum[i]; i++) {
		int digit, x, y, pixels;

		digit = spritenum[i] - '0';
		for (y=0; y<DIGITHEIGHT; y++) {
			pixels = digitlines[DIGITIND(digit,y)] >> DIGITSHR(digit);

			for (x=DIGITWIDTH-1; x>=0; x--) {
				if (pixels & 1)
					rows[thisy-6+y][thisx+x+i*(DIGITWIDTH+1)] = textcol;
				pixels >>= 1;
			}
		}

		lastnumx = thisx+(i+1)*(DIGITWIDTH+1);
	}

  }

  // write the image data
  for (i=0; i<height; i++) {
	uint32_t offset = putspriterow(data, flags, i, width, rows[thisy+i]+thisx);

	if (offset > dumped) dumped=offset;
  }

  *x = thisx;
  *y = thisy;

  return dumped;
}

//
// Write a sprite to the given PNG file, or the complete all-in-one data
//
uint32_t makepng(FILE *xml, FILE *png, uint8_t *data, uint32_t width, uint32_t height, uint32_t flags)
{
  png_structp png_ptr;
  png_infop info_ptr;
  png_colorp palette;
  png_byte *row;
  png_color_16 backgnd;

  int i, j, ccolind;
  uint32_t dumped;

  dumped = 0;

#ifdef __windows
	// make sure the CR<->CRLF translation is off
  if (setmode(fileno(xml), O_BINARY) == 1) die("Can't switch to binary file mode");
#endif

	// create the png structures
  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr) die("Error creating png write struct");

  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
	png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
	die("Error creating png info struct");
  }

	// error handlingg
  if (setjmp(png_jmpbuf(png_ptr))) {
		// if we get here, we had a problem writing the file
	png_destroy_write_struct(&png_ptr, &info_ptr);
	die("Error while writing png data");
  }

  if (embedpng) {
		// use our own write functions to write the data to the
		// xml file instead
	png_set_write_fn(png_ptr, xml, xml_write, xml_flush);
  } else if (allinone) {
	png_init_io(png_ptr, png);
	width = 800;
	height = numrows;
  } else {
	png_init_io(png_ptr, png);
  }

	// set image dimensions
  png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_PALETTE,
    PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	// set the palette
  palette = (png_colorp)png_malloc(png_ptr, PNG_MAX_PALETTE_LENGTH * sizeof (png_color));
  if (!palette) die("Not enough memory for %s, wanted %d", "palette", PNG_MAX_PALETTE_LENGTH * sizeof (png_color));

  for (i=0; i<256; i++) {
	palette[i].red = locopalette[i][0];
	palette[i].green = locopalette[i][1];
	palette[i].blue = locopalette[i][2];
  }

	// company colours, set which colour to use
  ccolind = (palettetype&~1)*6-2;
  if (ccolind > 0) {
	for (i=0; i<12; i++) {
		palette[companycolours[i]].red = locopalette[ccolind+i][0];
		palette[companycolours[i]].green = locopalette[ccolind+i][1];
		palette[companycolours[i]].blue = locopalette[ccolind+i][2];
	}

		// disable this colour in non-company-colour range?
	if (palettetype & 1) {
		for (i=0; i<12; i++) {
			palette[ccolind+i].red = maskoutcol[0];
			palette[ccolind+i].green = maskoutcol[1];
			palette[ccolind+i].blue = maskoutcol[2];
		}
	}
  }

  png_set_PLTE(png_ptr, info_ptr, palette, PNG_MAX_PALETTE_LENGTH);

	// set the gamma
  png_set_gAMA(png_ptr, info_ptr, 1.4);

	// set the transparent colour
  png_set_tRNS(png_ptr, info_ptr, &trnscol, 1, NULL);

	// and the background colour
  backgnd.red = backgnd.green = backgnd.blue = 65535;
  png_set_bKGD(png_ptr, info_ptr, &backgnd);

	// write the png header
  png_write_info(png_ptr, info_ptr);

  if (embedpng) {
  } else if (allinone) {
	for (i=0; i<height; i++) {
		png_write_row(png_ptr, rows[i]);
		my_free(rows[i]);
	}
	my_free(rows);

	allocrows = 0;
	numrows = 0;
	nextx = 65535;
	lastnumx = -200;
	nexty = 8;
  } else {
	row = my_alloc(width, "row");

		// write the image line by line
	for (i=0; i<height; i++) {
		uint32_t offset = putspriterow(data, flags, i, width, row);

		if (offset > dumped) dumped=offset;

		png_write_row(png_ptr, row);
	}

	my_free(row);
  }

	// done writing, clean up
  png_write_end(png_ptr, info_ptr);
  png_free(png_ptr, palette);
  png_destroy_write_struct(&png_ptr, &info_ptr);

#ifdef __windows
	// switch back to text mode for the remainder of the XML
  if (setmode(fileno(xml), O_TEXT) == 1) die("Can't switch to text file mode");
#endif

  return dumped;
}

//
// Read a png file, return its data and properties
//
png_byte **readpng(const char *name, uint32_t *width, uint32_t *height)
{
  png_structp png_ptr;
  png_infop info_ptr;
  int bit_depth, color_type, interlace_type;
  FILE *png;
  int i;
  png_byte **rows, **copy;
  png_colorp palette;

  png = fopen(name, "rb");
  if (!png) die("Can't read %s: %s", name, strerror(errno));

  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr) {
	fclose(png);
	die("Error creating png read struct");
  }

  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
	fclose(png);
	png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
	die("Error creating png info struct");
  }

  if (setjmp(png_jmpbuf(png_ptr)))
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    fclose(png);
	die("Error while reading png data");
  }

  png_init_io(png_ptr, png);

  png_read_png(png_ptr, info_ptr, 0, (png_infopp)NULL);

  *width = png_get_image_width(png_ptr, info_ptr);
  *height = png_get_image_height(png_ptr, info_ptr);

  if (png_get_bit_depth(png_ptr, info_ptr) != 8)
	die("Unsupported PNG bit depth in %s", name);

  bgcol = 255;
  if (png_get_color_type(png_ptr, info_ptr) & PNG_COLOR_MASK_PALETTE) {
	png_get_PLTE(png_ptr, info_ptr, &palette, &i);
	if ( (palette[255].blue != 255) )
		bgcol = 0;
  }

  rows = png_get_rows(png_ptr, info_ptr);

  // make a copy of the data, because libpng will free it...
  copy = my_alloc(*height * sizeof(copy[0]), "copy");
  for (i=0; i<*height; i++) {
	copy[i] = my_alloc(*width, "rowcopy");
	memcpy(copy[i], rows[i], *width);
  }

  // clean up
  png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

  fclose(png);

  return copy;
}

uint32_t makesprite(png_byte **rows, uint8_t **data, int flags, int x, int y, uint32_t width, uint32_t height)
{
  int i;
  uint8_t *sprite;
  uint32_t maxsize, size;

  maxsize = height*width*3 + 2*height + 32;
  sprite = my_alloc(maxsize, "sprite");

  size = 0;
  if (flags & 4)
	size = height*2;

  for (i=0; i<height; i++) {
	size = getspriterow(sprite, size, flags, i, width, rows[i+y]+x);
	if (size >= maxsize) die("Sprite buffer overflow");
  }

  sprite = my_realloc(sprite, size, maxsize, "sprite");
  *data = sprite;

  return size;
}

void pngcleanup(png_byte **rows, int height)
{
  int i;

  for (i=0; i<height; i++)
	my_free(rows[i]);
  my_free(rows);
}
