//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// xml.c - read xml file using expat and parse its contents into an object
//	   data structure
//

#include <stdio.h>
#include <string.h>
#include <expat.h>
#include <stdint.h>
#include <sys/stat.h>

#include "common.h"
#include "codec.h"
#include "options.h"
#include "png.h"

//
// Writes a zero-terminated character string (minlen=0) or general
// byte array (length given by minlen), making sure to encode the XML
// meta characters
//
uint32_t putxmlstr(FILE *f, const char *buffer, uint32_t minlen)
{
  uint32_t i;

  if (!minlen) minlen = strlen(buffer);

  for (i=0; i<minlen; i++)
	switch (buffer[i]) {
		case '"': fputs("&quot;", f); break;
		case '<': fputs("&lt;", f); break;
		case '>': fputs("&gt;", f); break;
		case '&': fputs("&amp;", f); break;
		case '\\': fputs("\\\\", f); break;
		default: 
			if (buffer[i] < 32) {
				fprintf(f, "\\%02X", buffer[i]);
			}
			else {
				fputc(buffer[i], f);
			} break;
	}
  return minlen;
}

//
// Make sure the array described by inf is large enough to support
// writing an entry at ofs
//
// If ofs=-1, append one element to the end of the array
//
// Returns the (if necessary) modified ofs
//
uint32_t checklen(arrayinf *inf, int32_t ofs)
{
  uint32_t oldsize = inf->size;

  if (ofs<0) ofs=inf->ent;
  if (ofs>=inf->ent) inf->ent=ofs+1;

  if (inf->size >= (ofs+1)*inf->entsize) return ofs;

  while (inf->size < (ofs+1)*inf->entsize)
	inf->size += 128;

  if (oldsize)
	*(inf->data) = my_realloc(*(inf->data), inf->size, oldsize, "inf->data");
  else
	*(inf->data) = my_alloc(inf->size, "inf->data");

  return ofs;
}

//
// Append an entry to an array of pointers
//
void addarrent(arrayinf *inf, void *ent)
{
  uint32_t ofs = checklen(inf, -1);
  if ((*inf->data)[ofs]) my_free((*inf->data)[ofs]);
  (*inf->data)[ofs] = ent;
}

//
// Set an entry in an array of pointers
//
void setarrent(arrayinf *inf, int32_t ofs, void *ent)
{
  checklen(inf, ofs);
  if ((*inf->data)[ofs]) my_free((*inf->data)[ofs]);
  (*inf->data)[ofs] = ent;
}

// Definition of an element handler function
typedef void elemproc(xmlinf *inf, elemid id, int depth);

// Definition of an element
struct s_elem {
		// Element name
	const char *name;

		// Element id, and the possible containers it may be defined
		// in; if only one container is possible set both
		// container1 and container2 to the same ID
	elemid id, container1, container2, container3;

		// Handlers for the start and end tags of the element
	elemproc *start, *end;

		// Whether we want and need character data for this element
	int haschardata;
};
typedef struct s_elem elem;
elem elemlist[];	// Definition below


//
// Parse an attribute name
//
// Name can be of the form field, field_2A, field_2A[15] or field[15]
//
// This will return the name with the number and index removed
// (i.e. "field" in all the above cases), and the field number and array
// index in the parameters
//
void parsename(char **name, int *field, int *arrind)
{
  char *f, *a;

  *arrind = -1;
  *field = -1;

  a = strchr(*name, '[');
  f = strchr(*name, '_');

  if (a) {
	a[0] = 0;
	*arrind = strtol(a+1, NULL, 0);
  };
  if (f) {
	f[0] = 0;
	*field = strtol(f+1, NULL, 16);
  }
}

char *findalias(char *name)
{
  int i;

  for (i=0; namealiases[i][0]; i++)
	if (!strcmp(name, namealiases[i][0]))
		return (char *) namealiases[i][1];

  return NULL;
}

//
// Find the index of the variable whose name is stored in the given attribute
//
void findvar(xmlinf *inf, int depth, attrid attr, eleminf **elem, varinf **vars,
		int *field, int *num)
{
  char *name;
  int i;

  *elem = inf->elems[depth];
  name = (*elem)->atts[attr].str;
  if (!name) die("Element has no name!");

  parsename(&name, field, num);

  *vars = inf->elems[depth-1]->desc.vars;

  if (*field != -1) return;

  do {
	for (i=0; (*vars)[i].ofs != -1; i++)
		if (!strcmp(name, (*vars)[i].name)) {
			*field = (*vars)[i].ofs;
			return;
		}

	name = findalias(name);
  } while (name);

  die("Unknown variable name %s", (*elem)->atts[attr].str);
}

//
// Find the bit whose name is stored in the given attribute
//
int findbit(xmlinf *inf, int depth, attrid attr, eleminf **elem)
{
  char *name;
  int id, ind;
  varinf *vars;

  *elem = inf->elems[depth-1];
  name = inf->elems[depth]->atts[attr].str;
  if (!name) die("Element has no name!");

  parsename(&name, &id, &ind);

  if (id != -1) return id;

  if (!(*elem)->desc.flags)
	die("No bit flag information available");

  do {
	for (id=0; (*elem)->desc.flags[id]; id++)
		if (!strcmp((*elem)->desc.flags[id], name))
			return id;
	name = findalias(name);
  } while(name);

  die("Unknown bit name %s", inf->elems[depth]->atts[attr].str);
  return 0;	// to make the compiler happy
}

//
// Find the auxdata from the name
//
int findaux(xmlinf *inf, int depth, int *id, auxdesc *aux)
{
  int nameind;

  char *name = inf->elems[depth]->atts[at_name].str;

  parsename(&name, &nameind, id);

  if (nameind != -1)
	return nameind;

  if (!aux)
	die("No aux description!");

  do {
	for (nameind=0; aux[nameind].name; nameind++)
		if (!strcmp(aux[nameind].name, name))
			return nameind;
	name = findalias(name);
  } while(name);

  die("Unknown auxdata name %s", inf->elems[depth]->atts[at_name].str);
}

//
// Set the offset of the data which this element refers to
//
void setdata(eleminf *elem, eleminf *parent, int size, int ofs, int num)
{
  elem->data = ((uint8_t*)parent->data)+ofs+size*num;
}

//
// *****************************
//   Element handler functions
// *****************************
//
// These are called for start and end tags of each element, though not
// all elements have handlers for both.
//
// Typically, container elements like <structure> only handle the start
// tag to set up the sub-structure, and actual data elements like
// <variable> or <bit> only handle the end tag to store the data
//
// The naming convention is st_foo and en_foo for start and end tag
// handlers of the <foo> element.
//


// Null handler, nothing needs to be done
// (simpler than checking for NULL each time)
void st_null(xmlinf *inf, elemid id, int depth)
{
}

varinf auxvars[] = {
	{ 0x00, 0, 0, "" },
	{ -1 },
};

void st_auxdata(xmlinf *inf, elemid id, int depth)
{
  int num, size, nameind, ind;
  uint32_t len, ssize;
  auxdatainf *auxdata;
  char *name;
  eleminf *elem = inf->elems[depth];
  varinf *vars;
  auxdesc *auxinf;

  auxinf = objclasses[inf->class&0x3f].aux;

  name = elem->atts[at_name].str;
  if (!name) {
	nameind = -1;
	ind = elem->atts[at_id].u32;
  } else {
	nameind = findaux(inf, depth, &ind, auxinf);
	if (ind == -1) ind = 0;
  }

  num = elem->atts[at_num].u32;
  size = elem->atts[at_size].u32;
  len = num*size;

  auxdata = my_alloc(len+sizeof(auxdatainf), "auxdata");
  auxdata->nameind = nameind;
  auxdata->id = ind;
  auxdata->num = num;
  auxdata->size = size;
  auxdata->len = len;
  auxdata->type = elem->atts[at_type].u32;

  addarrent(&inf->auxdata_i, auxdata);

  auxvars[0].size = size;

  vars = auxinf[nameind].vars;
  if (!vars) vars = auxvars;

  ssize = structsize(vars);
  if (ssize != num * size) {
	vars[0].num = 1;
	ssize = structsize(vars);
	vars[0].num = num * size / ssize;
	if (vars[0].num * ssize != size * num)
		die("%s size %d*%d != %d*%d", name, ssize, vars[0].num, size, num);
  }

  elem->data = auxdata->data;
  elem->desc.vars = vars;
}

void en_bit(xmlinf *inf, elemid id, int depth)
{
  int bit, value, size;
  eleminf *elem;

  bit = findbit(inf, depth, at_name, &elem);

  value = strtol(inf->curvalue, NULL, 0);

  size = inf->elems[depth-1]->atts[at_size].u32;
  if (bit > size*8)
	die("Trying to set bit %d beyond size %d", bit, size);

  if (value)
	*(uint32_t*) elem->data |= (1 << bit);
}

void st_bitmask(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  varinf *vars;
  int var, size, ofs, num;

  findvar(inf, depth, at_name, &elem, &vars, &ofs, &num);
  if (num == -1) num=0;

  size = elem->atts[at_size].u32;
  if (!size)
	die("Structure '%s' has no size", elem->atts[at_name].str);

	// set up which bits subsequent <bit> tags will set
  setdata(elem, inf->elems[depth-1], size, ofs, num);

	// plus the flag descriptions
  elem->desc.flags = NULL;
  for (var=0; vars[var].ofs != -1; var++)
	if (vars[var].ofs == ofs)
		elem->desc.flags = vars[var].flags;
  if (!elem->desc.flags)
	die("Unknown bitmask");
}

void st_cargo(xmlinf *inf, elemid id, int depth)
{
  int num;

  num = inf->elems[depth]->atts[at_num].u32;
  if (num>1) die("Invalid capacity number %d", num);

  if (inf->capsize[num]) die("Duplicate capacity number %d", num);

	// store the capacity, and prepare structure for
	// subsequent <cargotype> tags, if any
	//
	// format of the data is
	// <capacity> [<cargotype> <byte1> <byte2>]... FF FF
  inf->capsize[num] = 128;
  inf->caps[num] = my_alloc(inf->capsize[num], "caps");
  inf->caps[num][0] = inf->elems[depth]->atts[at_capacity].u32;
  *(uint16_t*) (inf->caps[num]+1) = 0xffff;
  inf->caplen[num] = 3;
}

void en_cargotype(xmlinf *inf, elemid id, int depth)
{
  int num, len, ofs;

  num = inf->elems[depth-1]->atts[at_num].u32;

  ofs = inf->caplen[num];
  inf->caplen[num] += 3;
  while (inf->capsize[num] < inf->caplen[num]) {
	inf->caps[num] = my_realloc(inf->caps[num], inf->capsize[num]+128, inf->capsize[num], "caps");
	inf->capsize[num] += 128;
  }

	// append the cargo type
  inf->caps[num][ofs-2] = inf->elems[depth]->atts[at_id].u32;
  *(uint16_t*) (inf->caps[num]+ofs-1) = strtol(inf->curvalue, NULL, 0);
  *(uint16_t*) (inf->caps[num]+ofs+1) = 0xffff;
}

void st_chunk(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem, *parent;

  elem = inf->elems[depth];
  parent = inf->elems[depth-1];

  inf->compression = elem->atts[at_compression].u32;

	// copy data description of surrounding <object> tag
  elem->data = parent->data;
  elem->desc.flags = parent->desc.flags;
}

void en_description(xmlinf *inf, elemid id, int depth)
{
  int ofs;
  descinf *desc;
  eleminf *elem;

  desc = my_alloc(inf->curvaluelen+1+sizeof(descinf), "desc");

  elem = inf->elems[depth];
  desc->num = elem->atts[at_num].u32;
  desc->len = inf->curvaluelen+1;
  desc->lang = elem->atts[at_language].u32;

  memcpy(desc->data, inf->curvalue, inf->curvaluelen);

  addarrent(&inf->descdata_i, desc);
}

void st_object(xmlinf *inf, elemid id, int depth)
{
  char *name;
  eleminf *elem = inf->elems[depth];

  inf->class = elem->atts[at_class].u32;
  inf->subclass = elem->atts[at_subclass].u32;

  memset(inf->name, ' ', 8);

  name=elem->atts[at_name].str;
  if (strlen(name)>8)
	die("Object name '%s' too long", name);

  memcpy(inf->name, name, strlen(name));	// padded; no terminating zero

  if ((inf->class&0x3f) < NUMOBJCLASSES)
	elem->desc.vars = objclasses[inf->class&0x3f].vars;

  if (!elem->desc.vars)
	die("Cannot parse class %X yet", inf->class);

	// pre-defined the array length needed for the object data
  checklen(&inf->objdata_i, objclasses[inf->class & 0x3f].size);
  elem->data = inf->objdata;
}

void en_pngfile(xmlinf *inf, elemid id, int depth)
{
  uint8_t *spritedata;
  uint32_t width, height, size;
  spriteinf *sprite;
  png_byte **rows;

  sprite = inf->sprites[inf->cursprite];

  if (!sprite->flags)
	die("No bits set for sprite id %d", inf->cursprite);

  rows = readpng(inf->curvalue, &width, &height);
  size = makesprite(rows, &spritedata, sprite->flags, 0, 0, width, height);
  pngcleanup(rows, height);

  setarrent(&inf->spritedata_i, inf->cursprite, spritedata);
  sprite->width = width;
  sprite->height = height;
  sprite->spritesize = size;

	// prevent <bit> tags after we already have the data
  inf->elems[depth-1]->desc.flags = NULL;
}

void en_pngimage(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  uint8_t *spritedata;
  uint32_t x, y, width, height, size;
  spriteinf *sprite;

  elem = inf->elems[depth];

  sprite = inf->sprites[inf->cursprite];

  if (!sprite->flags)
	die("No bits set for sprite id %d", inf->cursprite);

  if (inf->curpngfile && strcmp(inf->curpngfile, inf->curvalue)) {
	my_free(inf->curpngfile);
	pngcleanup(inf->curpngrows, inf->curpngheight);
  }
  if (!inf->curpngfile) {
	inf->curpngfile = my_strdup(inf->curvalue, "curpngfile");
	inf->curpngrows = readpng(inf->curpngfile, &inf->curpngwidth, &inf->curpngheight);
  }

  x = elem->atts[at_xofs].u32;
  y = elem->atts[at_yofs].u32;
  width = elem->atts[at_width].u32;
  height = elem->atts[at_height].u32;

  if ( (x+width > inf->curpngwidth) || (y+height > inf->curpngheight) )
	die("Sprite %d outside of PNG file", inf->cursprite);

  size = makesprite(inf->curpngrows, &spritedata, sprite->flags, x, y, width, height);

  setarrent(&inf->spritedata_i, inf->cursprite, spritedata);
  sprite->width = width;
  sprite->height = height;
  sprite->spritesize = size;

	// prevent <bit> tags after we already have the data
  inf->elems[depth-1]->desc.flags = NULL;
}

void st_sprite(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem = inf->elems[depth];
  spriteinf *sprite = my_alloc(sizeof(spriteinf), "sprite");

  inf->cursprite = elem->atts[at_id].u32;
  elem->data = &sprite->flags;
  elem->desc.flags = spriteflags;
  elem->atts[at_size].u32 = 4;
  sprite->xofs = elem->atts[at_xofs].u32;
  sprite->yofs = elem->atts[at_yofs].u32;

  setarrent(&inf->sprites_i, inf->cursprite, sprite);
}

void en_sprite(xmlinf *inf, elemid id, int depth)
{
  spriteinf *old, *new;

  new = inf->sprites[inf->cursprite];
  if (new->flags & 0x40) {
	old = inf->sprites[inf->cursprite-1];
	new->width = old->width;
	new->height = old->height;
	new->spritesize = old->spritesize;
  } else if (!inf->spritedata[inf->cursprite])
	die("No sprite data for sprite id %d", inf->cursprite);
}

void en_string(xmlinf *inf, elemid id, int depth)
{
  int ofs;
  string *str;

  str = my_alloc(inf->curvaluelen+1+sizeof(string), "string");

  str->id = inf->curstrtableid;
  str->num = inf->elems[depth]->atts[at_id].u32;
  str->type = inf->elems[depth]->atts[at_type].u32;
  str->len = inf->curvaluelen+1;

  memcpy(str->str, inf->curvalue, inf->curvaluelen);

  addarrent(&inf->strtable_i, str);
}

void st_stringtable(xmlinf *inf, elemid id, int depth)
{
  inf->curstrtableid = inf->elems[depth]->atts[at_id].u32;
}

void st_structure(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  varinf *vars;
  int var, ofs, num, size;

  findvar(inf, depth, at_name, &elem, &vars, &ofs, &num);
  if (num == -1) num=0;

  size = elem->atts[at_size].u32;
  if (!size)
	die("Structure '%s' has no size", elem->atts[at_name].str);

  setdata(elem, inf->elems[depth-1], size, ofs, num);

  for (var=0; vars[var].ofs != -1; var++)
	if (vars[var].ofs == ofs)
		elem->desc.vars = vars[var].structvars;

  if (!elem->desc.vars)
	die("Structure '%s' has no definition", elem->atts[at_name].str);
}

void en_stub(xmlinf *inf, elemid id, int depth)
{
  uint8_t *spritedata, col;
  uint32_t width, height, size;
  spriteinf *sprite;

  sprite = inf->sprites[inf->cursprite];

  if (!sprite->flags)
	die("No bits set for sprite id %d", inf->cursprite);

  spritedata = my_alloc(16, "spritedata");
  col = strtol(inf->curvalue, NULL, 0);
  size = getspriterow(spritedata, 2, sprite->flags, 0, 1, &col);

  setarrent(&inf->spritedata_i, inf->cursprite, spritedata);
  sprite->width = 1;
  sprite->height = 1;
  sprite->spritesize = size;

	// prevent <bit> tags after we already have the data
  inf->elems[depth-1]->desc.flags = NULL;
}

//
// list of all support <useobject> types
//
// NOTE: Any changes here must be synchronized with the en_useobjid enum in locodef.h
//
const char *useobjnames[ob_LASTVALUE] = {
	"tracktype", "trackmod", "visualeffect", "wakeeffect",
	"rackrail", "compatible", "startsnd", "soundeffect",
	"signal", "tunnel", "bridge", "station", "cliff",
	"roadmod", "produces", "accepts", "fence", "cargo",
	"default",
};

void en_useobject(xmlinf *inf, elemid id, int depth)
{
  useobjid type;
  eleminf *elem;
  char *name;
  int field, num;
  useobjinf *objinf = my_alloc(sizeof(useobjinf), "objinf");

  elem = inf->elems[depth];
  objinf->class = elem->atts[at_class].u32;

  name = elem->atts[at_desc].str;
  parsename(&name, &field, &num);
  if (num == -1) num=0;

  for (type=0; type<ob_LASTVALUE; type++)
	if (!strcmp(name, useobjnames[type]))
		break;
  if (type == ob_LASTVALUE) die("Unknown object %s", elem->atts[at_desc].str);

  objinf->type = type;
  objinf->num = num;

  if (inf->curvaluelen > 8)
	die("Object name '%s' too long", inf->curvalue);

	// assembly the data
	// format is
	// <class> 00 00 00 <name> 00 00 00 00
	// where <name> is the name space-padded to eight bytes
  objinf->obj[0] = objinf->class;
  memset(objinf->obj+4, ' ', 8);
  strncpy(objinf->obj+4, inf->curvalue, 8);

  addarrent(&inf->useobj_i, objinf);
}

void en_variable(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  varinf *vars;
  int ofs, num, size;
  uint32_t value;

  findvar(inf, depth, at_name, &elem, &vars, &ofs, &num);
  if (num == -1) num=0;

  size = elem->atts[at_size].u32;
  setdata(elem, inf->elems[depth-1], size, ofs, num);

  value = strtol(inf->curvalue, NULL, 0);
  switch(size) {
	case 1: *(uint8_t*) elem->data = value; break;
	case 2: *(uint16_t*) elem->data = value; break;
	case 4: *(uint32_t*) elem->data = value; break;
	default: die("Invalid size %d", size);
  }
}

// <unknown> is a <variable> whose meaning is unknown
void en_unknown(xmlinf *inf, elemid id, int depth)
{
  en_variable(inf, id, depth);
}

// null handler for the end tag of a container
void en_updepth(xmlinf *inf, elemid id, int depth)
{
}

// wav file
void en_wavfile(xmlinf *inf, elemid id, int depth)
{
  FILE *wav;
  uint8_t *wavdata;
  uint32_t width, height, size;
  spriteinf *sprite;
  struct stat st;

  wav = fopen(inf->curvalue, "rb");
  if (!wav) die("Cannot read %s: %s", inf->curvalue, strerror(errno));
  if (fstat(fileno(wav), &st)) die("Can't stat %s: %s", inf->curvalue, strerror(errno));

  wavdata = my_alloc(st.st_size, "wavdata");
  fread(wavdata, 1, st.st_size, wav);
  fclose(wav);

  // store it like sprite data for simplicity
  setarrent(&inf->spritedata_i, 0, wavdata);

  sprite = my_alloc(sizeof(spriteinf), "sprite");
  sprite->width = inf->elems[depth]->atts[at_size].u32;
  sprite->height = -1;
  sprite->spritesize = st.st_size;
  setarrent(&inf->sprites_i, 0, sprite);
}


// define handlers for a single value, a simple container, and a sub-structure
#define XMLVALUE(name,cont1,cont2,cont3) \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, el_ ## cont3, st_null, en_ ## name, 1 }
#define XMLVALUESTR(name,cont1,cont2,cont3) \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, el_ ## cont3, st_null, en_ ## name, 2 }
#define XMLCONT(name,cont1,cont2,cont3)  \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, el_ ## cont3, st_ ## name, en_updepth, 0 }
#define XMLSTRUC(name,cont1,cont2,cont3) \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, el_ ## cont3, st_ ## name, en_ ## name, 0 }

//
// list of all supported XML elements
//
// NOTE: Any changes here must be synchronized with the en_elemid enum in locodef.h
//
elem elemlist[] = {
	XMLVALUE(bit, bitmask, sprite, sprite),
	XMLCONT(bitmask, chunk, structure, auxdata),
	XMLCONT(cargo, chunk, chunk, chunk),
	XMLVALUE(cargotype, cargo, cargo, cargo),
	XMLCONT(chunk, object, object, object),
	XMLVALUESTR(description, chunk, chunk, chunk),
	XMLCONT(object, document, document, document),
	XMLVALUESTR(pngfile, sprite, sprite, sprite),
	XMLSTRUC(sprite, chunk, chunk, chunk),
	XMLCONT(structure, chunk, structure, auxdata),
	XMLVALUE(unknown, chunk, structure, auxdata),
	XMLVALUE(useobject, chunk, chunk, chunk),
	XMLVALUE(variable, chunk, structure, auxdata),
	XMLVALUE(stub, sprite, sprite, sprite),
	XMLCONT(auxdata, chunk, chunk, chunk),
	XMLCONT(stringtable, chunk, chunk, chunk),
	XMLVALUESTR(string, stringtable, stringtable, stringtable),
	XMLVALUE(wavfile, chunk, chunk, chunk),
	XMLVALUESTR(pngimage, sprite, sprite, sprite),
	{ NULL },
};

//
// list of all supported XML attributes
//
//
// NOTE: Any changes here must be synchronized with the en_attrid enum in locodef.h
//
const char *attname[at_LASTVALUE] = {
	"capacity", "class", "compression", "id", "type",
	"language", "num", "size", "subclass", "xofs", "yofs",
        "width", "height",

	"",	// at_STRINGVALUE is not a real attribute

	"name", "desc",
};

//
// called by expat for each start tag
//
static void XMLCALL startElement(void *userdata, const char *name, const char **atts)
{
  int i, j, cont1, cont2, cont3, depth;
  eleminf *elem;
  xmlinf *inf = userdata;

  // search for the element name
  for (i=0; elemlist[i].name; i++) {
	if (strcmp(name, elemlist[i].name)) continue;

	if (i != elemlist[i].id) die("Inconsistent elemlist entry %s", name);

	// check that it has the right container
	depth=inf->elems_i.ent;
	j = inf->elems[depth-1]->id;
	cont1 = elemlist[i].container1;
	cont2 = elemlist[i].container2;
	cont3 = elemlist[i].container3;

	if ( (j != cont1) && (j != cont2) && (j != cont3) )
		die("Element %s is inside element %s but can only be a child of %s%s%s%s%s",
			name, elemlist[j].name, elemlist[cont1].name,
			cont1 != cont2 ? " or " : "",
			cont1 != cont2 ? elemlist[cont2].name : "",
			cont2 != cont3 ? " or " : "",
			cont2 != cont3 ? elemlist[cont3].name : "");

	inf->chardata = elemlist[i].haschardata;

	// define the element structure
	elem = my_alloc(sizeof(eleminf), "elem");
	elem->id = i;

	// parse the attributes
	while (*atts) {
		for (j=0; j<=at_LASTVALUE; j++) {
			if (j == at_LASTVALUE)
				die("Unknown attribute '%s' for element '%s'", name, atts[0]);

			if (strcmp(atts[0], attname[j]) == 0) {
				if (j < at_STRINGVALUE)
					elem->atts[j].u32 = strtol(atts[1], NULL, 0);
				else
					elem->atts[j].str = my_strdup(atts[1], "elem atts str");
				atts+=2;
				break;
			}
		}
	}

	// and append it to the end of the element stack
	addarrent(&inf->elems_i, elem);

	// finally call the start tag handler
	elemlist[i].start(inf, i, depth);
	return;
  }

  die("Unknown XML element %s", name);
}

// same for the end tag
static void XMLCALL endElement(void *userdata, const char *name)
{
  int i;
  elemid id;
  attrid j;
  eleminf *elem;
  xmlinf *inf = userdata;

  i = inf->elems_i.ent-1;
  elem = inf->elems[i];
  id = elem->id;

  // expat should catch <bla></foo> but better make sure :)
  if (strcmp(elemlist[id].name, name))
	die("Incorrect end element %s", name);

  // did we need character data? if so make sure we didn't have an empty element
  if ((inf->chardata == 1) && !inf->curvaluelen)
	die("Element '%s' is empty", name);

  // call the end tag handler
  elemlist[id].end(inf, id, i);

  // and free the element structure
  for (j=at_STRINGVALUE+1; j<at_LASTVALUE; j++)
	my_free(elem->atts[j].str);	// only free string attributes

  // finally remove it from the element stack
  my_free(inf->elems[i]);
  inf->elems_i.ent--;

  my_free(inf->curvalue);
  inf->curvaluelen = 0;
  inf->curvaluesize = 0;
  inf->chardata = 0;	// by default we don't want character data
}

//
// Character data handler, called for every byte that isn't a tag
//
// We skip everything between container tags and their sub-element tags
// but keep the data between e.g. <variable> and </variable>
//
// Note: the string s is in UTF-8 format, regardless of the encoding of
//	 the XML document
//
static void XMLCALL characterdata(void *userdata, const XML_Char *s, int utf8len)
{
  xmlinf *inf = userdata;
  uint32_t i, len, newlen, oldlen, newsize;
  uint8_t c;
  char hex[3];

  if (!inf->chardata) return;

  // find the string length in characters (not bytes)
  // we only support characters < 255 (i.e. those with UTF-8 encoding of up
  // to two bytes)
  len = 0;
  for (i=0; i<utf8len; i++) {
	c = s[i];

	// check if it's a malformed UTF-8 sequence or a character > 255
	if ((c>0xe0) || ( (c>=0x80) && (c<0xc2) ) )
		die("Invalid character sequence %02X%02X", (uint8_t) s[i], (uint8_t) s[i+1]);
	if (c & 0x80) {
		c &= 0xf0;
		while (c & 0x80) {
			c <<= 1;
			i++;
		}
		i--;
	}
	if ((inf->chardata == 2) && (c == '\\'))
		if (s[++i] != '\\')
			i++;
	len++;
  }

  // prepare curvalue array to append the string
  oldlen = inf->curvaluelen;
  newlen = oldlen + (uint32_t) len;
  newsize = inf->curvaluesize;

  if (newsize <= newlen+1) {
	while (newsize <= newlen+1)
		newsize += 128;

	if (oldlen)
		inf->curvalue = my_realloc(inf->curvalue, newsize, oldlen, "curvalue");
	else
		inf->curvalue = my_alloc(newsize, "curvalue");
	inf->curvaluesize = newsize;
  }

  // append the string character by character is ISO-8859-1 encoding
  len = 0;
  for (i=0; i<utf8len; ) {
	c = s[i++];

	// decode UTF-8
	if (c & 0x80)
		c = ((c & 0x3f) << 6) + (s[i++] & 0x3f);
	if ((inf->chardata == 2) && (c == '\\')) {
		hex[0] = s[i++];

		if (hex[0] != '\\') {
			hex[1] = s[i++];
			hex[2] = 0;
			c = strtol(hex, NULL, 16);
		}
	}

	inf->curvalue[oldlen+len++] = c;
  }
  inf->curvalue[newlen] = 0;
  inf->curvaluelen = newlen;
}


//
// ***********************************
//   Object data structure functions
// ***********************************
//
// These functions help the class-specific xml assemblers to append data
// to the end of the object data array to compose the whole .dat structure
//

// Append an arbitrary sized array
void addval(xmlinf *inf, uint8_t *val, uint32_t len)
{
  uint32_t ofs = inf->objdata_i.ent;

  checklen(&inf->objdata_i, ofs+len-1);
  memcpy(inf->objdata+ofs, val, len);
}

// Append a 4-byte value
void add32(xmlinf *inf, uint32_t val)
{
  uint32_t ofs = inf->objdata_i.ent;

  checklen(&inf->objdata_i, ofs+3);
  *(uint32_t*) (inf->objdata+ofs) = val;
}

// Append a 2-byte value
void add16(xmlinf *inf, uint16_t val)
{
  uint32_t ofs = inf->objdata_i.ent;

  checklen(&inf->objdata_i, ofs+1);
  *(uint16_t*) (inf->objdata+ofs) = val;
}

// Append a single byte
void add8(xmlinf *inf, uint8_t val)
{
  uint32_t ofs = inf->objdata_i.ent;

  checklen(&inf->objdata_i, ofs);
  inf->objdata[ofs] = val;
}

// Append a set of translations
void adddesc(xmlinf *inf, int num)
{
  int i;

  for (i=0; i<inf->descdata_i.ent; i++)
	if (inf->descdata[i]->num == num) {
		add8(inf, inf->descdata[i]->lang);
		addval(inf, inf->descdata[i]->data, inf->descdata[i]->len);
	}
  add8(inf, 0xff);
}

// Append an external object dependence
int addobj(xmlinf *inf, int num, useobjid id, int32_t *classes)
{
  int i, j, classok;

  for (i=0; i<inf->useobj_i.ent; i++)
	if ( (inf->useobj[i]->type == id) && (inf->useobj[i]->num == num) ) {

		classok = 0;
		for (j=0; classes[j] != -1; j++)
			if ( (classes[j] == inf->useobj[i]->class) || (classes[j] == 0x100) )
				classok = 1;

		if (!classok)
			die("Not correct class for %s: %02X", useobjnames[id], inf->useobj[i]->class);

		addval(inf, inf->useobj[i]->obj, 16);
		return 1;
	}

  return 0;
}

// Append the list of sprites and their data
void addsprites(xmlinf *inf)
{
  uint32_t i, size, lastofs, ofs;

  add32(inf, inf->sprites_i.ent);

  size = 0;
  for (i=0; i<inf->sprites_i.ent; i++)
	if (!(inf->sprites[i]->flags & 0x40))
		size += inf->sprites[i]->spritesize;

  add32(inf, size);

  lastofs = ofs = 0;
  for (i=0; i<inf->sprites_i.ent; i++) {
	spriteinf *s=inf->sprites[i];

	if (s->flags & 0x40)
		ofs = lastofs;

	lastofs = ofs;
	add32(inf, ofs);
	add16(inf, s->width);
	add16(inf, s->height);
	add16(inf, s->xofs);
	add16(inf, s->yofs);
	add32(inf, s->flags);

	ofs += inf->sprites[i]->spritesize;
  }

  for (i=0; i<inf->sprites_i.ent; i++)
	if (!(inf->sprites[i]->flags & 0x40))
		addval(inf, inf->spritedata[i], inf->sprites[i]->spritesize);
}

int auxmatches(xmlinf *inf, auxdatainf *aux, int nameind, int id)
{
  return (
	   ( (aux->nameind == nameind) && (aux->id == id) )
		||
	   ( (aux->nameind == -1) && (aux->id == inf->nextauxid) )
  );
}

int countaux(xmlinf *inf, int nameind, int id)
{
  int j, num;

  num = -1;
  for (j=0; j<inf->auxdata_i.ent; j++)
	if (auxmatches(inf, inf->auxdata[j], nameind, id)) {
		num = inf->auxdata[j]->num;
		break;
	}
  return num;
}

// auxiliary data of unknown format
void addaux(xmlinf *inf, int nameind, int id)
{
  int i;
  int32_t val = -1;	// needs to be 32 bit because type may be 1, 2 or 4

  for (i=0; i<inf->auxdata_i.ent; i++)
	if (auxmatches(inf, inf->auxdata[i], nameind, id)) {
		addval(inf, inf->auxdata[i]->data, inf->auxdata[i]->len);
		if (inf->auxdata[i]->type)
			addval(inf, (uint8_t*) &val, inf->auxdata[i]->type);
		inf->nextauxid++;
		break;
	}
}

// string table
void addstrtable(xmlinf *inf, int id, uint32_t numofs, uint32_t dataofs)
{
  int i, j, num, rep;
  uint32_t ofs;

  num = 0;

  // set offset
  if (inf->objdata[numofs])
	*(uint16_t*) (inf->objdata+dataofs) = inf->objdata_i.ent;

	  // first count them
  for (i=0; i<inf->strtable_i.ent; i++)
	if (inf->strtable[i]->id == id)
		num++;

	// then set the offsets
  ofs = num*2;
  j = 0;
  for (i=0; i<num; i++) {
	rep = 0;
	while ( (inf->strtable[j]->id != id) || (inf->strtable[j]->num != i) ) {
		j++;
		if (j >= inf->strtable_i.ent) {
			rep++;
			if (rep > 2)
				die("Could not find string id %d num %d", id, i);
			j=0;
		}
	}
	add16(inf, ofs);
	ofs += inf->strtable[j]->len+1;
  }

	// and finally the actual strings
  j=0;
  for (i=0; i<num; i++) {
	while ( (inf->strtable[j]->id != id) || (inf->strtable[j]->num != i) ) {
		j++;
		if (j >= inf->strtable_i.ent) j=0;
	}
	addval(inf, inf->strtable[j]->str, inf->strtable[j]->len);
	add8(inf, inf->strtable[j]->type);
  }
}

void addsounds(xmlinf *inf)
{
  uint32_t len;
  int i;

  len = inf->sprites[0]->spritesize - 0x2c;

  add32(inf, 1);
  add32(inf, len+16+8+4); //+ (6));
  for (i=0; i<4; i++)
	add32(inf, 0);
  add32(inf, 1);
  add32(inf, 8);
  add32(inf, len-inf->sprites[0]->width);
  addval(inf, inf->spritedata[0]+0x14, 16);
  addval(inf, inf->spritedata[0]+0x2c, len);
}

// re-assembly the object class-specific data
void parsechunk(xmlinf *inf)
{
  int class, i, j, num, num2;
  int32_t *params;
  uint32_t dumped;
//  uint8_t zero;
//
//  zero = 0;

  class = inf->class & 0x3f;
  if ( (class >= NUMOBJCLASSES) || !objclasses[class].desc)
	die("Can't translate class %02X yet", class);

  if (objclasses[class].desc[0].type != desc_objdata)
	die("Object class has no objdata at start");

  for (i=0; objclasses[class].desc[i].type != desc_END; i++) {
	params = objclasses[class].desc[i].param;

	switch (objclasses[class].desc[i].type) {
		case desc_objdata:	// -
			if (i != 0)
				die("Object class has objdata not at start");
			break;

		case desc_lang:		// langid
			adddesc(inf, params[0]);
			break;

		case desc_useobj:	// num* type classes...
			for (j=0; ; j++) {
				dumped = 0;
				num = getnum(inf->objdata, &dumped, params[0]);

				if (loopescape(j,num)) break;

				if (!addobj(inf, j, params[1], params+2)) {
					if (num == 0xffffff) {
						add8(inf, 0xff);
						break;
					}
					die("Could not find object to use for %s[%d]", useobjnames[params[1]], num);
				}
			};
			break;

		case desc_auxdata:	// nameind numaux* size num*
		case desc_auxdatavar:	// nameind numaux* size type
			num = getnum(inf->objdata, NULL, params[1]);
			for (j=0; !loopescape(j,num); j++)
				addaux(inf, params[0], j);
			break;

		case desc_auxdatafix:	// nameind numaux* size numsize
			num = getnum(inf->objdata, NULL, params[1]);
			for (j=0; !loopescape(j,num); j++) {
				num2 = countaux(inf, params[0], j);
				addval(inf, (uint8_t*) &num2, params[3]);
				addaux(inf, params[0], j);
			}
			break;

		case desc_strtable:	// id num* ofs*
			addstrtable(inf, params[0], params[1], params[2]);
			break;

		case desc_cargo:	// num*
			num = getnum(inf->objdata, NULL, params[0]);
			for (j=0; !loopescape(j,num); j++) {
				uint8_t zero = 0;

				if (inf->capsize[j])
					addval(inf, inf->caps[j], inf->caplen[j]);
				else
					addval(inf, &zero, 1);
			}
			break;

		case desc_sprites:	// -
			addsprites(inf);
			break;

		case desc_sounds:	// -
			addsounds(inf);
			break;

		default:
			die("Unknown object description type %d", objclasses[class].desc[i].type);
	}
  }

  return;
}


//
// ****************************
//   XML parser main function
// ****************************
//
// Set up the XML parsing, parse the XML and call the object class-specific
// handler to assemble the rest of the object data
//

// initialize an arrayinf structure for all auto-sized arrays
#define MAKEARR(name) \
	parsedinfo->name ## _i.entsize = sizeof(parsedinfo->name[0]); \
	parsedinfo->name ## _i.data = (void*) &(parsedinfo->name);

// free all entries of an auto-sized array along with the array itself
void freearr(arrayinf *inf)
{
  int i;

  for (i=0; i<inf->ent; i++)
	my_free((*inf->data)[i]);
  my_free(*inf->data);
  *inf->data=NULL;
}

// the main parser function
uint32_t parsexml(uint8_t *data, uint32_t len, uint8_t **outdata, int *compression)
{
  XML_Parser parser;
  xmlinf *parsedinfo;
  eleminf *doc;
  uint32_t outlen;
  int i;

  // set up the xmlinf structure parsedinfo
  parsedinfo = my_alloc(sizeof(xmlinf), "parsedinfo");
  MAKEARR(objdata);
  MAKEARR(descdata);
  MAKEARR(useobj);
  MAKEARR(auxdata);
  MAKEARR(strtable);
  MAKEARR(sprites);
  MAKEARR(spritedata);
  MAKEARR(elems);

  // the top of the stack of elements is the notional "document" element
  doc = my_alloc(sizeof(eleminf), "doc");
  doc->id = el_document;
  addarrent(&parsedinfo->elems_i, doc);

  // setup the expat stuff and call the expat parser
  parser = XML_ParserCreate(NULL);
  XML_SetUserData(parser, parsedinfo);
  XML_SetElementHandler(parser, startElement, endElement);
  XML_SetCharacterDataHandler(parser, characterdata);
  if (XML_Parse(parser, data, len, 1) == XML_STATUS_ERROR)
	die("Error while parsing xml: %s at line %d",
	      XML_ErrorString(XML_GetErrorCode(parser)),
	      XML_GetCurrentLineNumber(parser));
  XML_ParserFree(parser);

  // this shouldn't happen
  if (parsedinfo->elems_i.ent != 1) die("Recursion not terminated");

  if ((parsedinfo->class & 0x3f) >= NUMOBJCLASSES)
	die("This can't happen");

  // assemble the rest of the object data
  parsedinfo->objdata_i.ent = objclasses[parsedinfo->class & 0x3f].size;

  parsechunk(parsedinfo);

  // at the beginning, insert the file header of class, subclass,
  // space-padded name and space for the checksum
  outlen = parsedinfo->objdata_i.ent;
  checklen(&parsedinfo->objdata_i, outlen+16);
  *outdata = parsedinfo->objdata;
  *compression = parsedinfo->compression;

  memmove(parsedinfo->objdata+16, parsedinfo->objdata, outlen);
  outlen += 16;

  parsedinfo->objdata[0] = parsedinfo->class;
  memcpy(parsedinfo->objdata+1, &parsedinfo->subclass, 3);
  memcpy(parsedinfo->objdata+4, parsedinfo->name, 8);
  memset(parsedinfo->objdata+12, 0, 4);

  // We don't free parsedinfo->objdata because we return that instead

  freearr(&parsedinfo->descdata_i);
  freearr(&parsedinfo->useobj_i);
  freearr(&parsedinfo->auxdata_i);
  freearr(&parsedinfo->strtable_i);
  my_free(parsedinfo->caps[0]);
  my_free(parsedinfo->caps[1]);
  freearr(&parsedinfo->sprites_i);
  freearr(&parsedinfo->spritedata_i);
  freearr(&parsedinfo->elems_i);
  if (parsedinfo->curpngfile) {
	my_free(parsedinfo->curpngfile);
	pngcleanup(parsedinfo->curpngrows, parsedinfo->curpngheight);
  }

  my_free(parsedinfo);

  return outlen;
}


// Exported: Encode the given xml file into a Locomotion .dat file
int LocoTool_encode(char *xmlname)
{
  int compression;
  uint32_t ofs, rlelen, chksum, encsize;
  struct stat st;

  FILE *xml, *dat;
  char *datname, *tmpname, *bakname;
  uint8_t *xmldata, *rle, *encdata;
  jmp_buf encode_err;

  // error handler, on error we return here with 0
  err_return = &encode_err;
  if (setjmp(*err_return)) {
	err_return = NULL;
	return 0;
  }

  setfilebase(xmlname);

  datname = makefilename(".dat");
  tmpname = makefilename(".tmp");
  bakname = makefilename(".bak");

  xml = fopen(xmlname, "rb");	// rt for real xml!!
  if (!xml) die("Can't read %s: %s", xmlname, strerror(errno));

  if (fstat(fileno(xml), &st)) die("Can't stat %s: %s", xmlname, strerror(errno));

  // slurp the whole xml file
  xmldata = my_alloc(st.st_size, "data");
  fread(xmldata, 1, st.st_size, xml);
  fclose(xml);

  // parse it
  encsize = parsexml(xmldata, st.st_size, &encdata, &compression);
  my_free(xmldata);

  // write the file header after calculating the checksum
  dat = fopen(tmpname, "w+b");
  chksum = LocoTool_checksum(encdata, encsize);
  *(uint32_t*) (encdata+12) = chksum;
  fwrite(encdata, 1, 16, dat);

  // For debugging the object data assembly functions
#ifdef DEBUGDMP
  {
    FILE *f;
    char *dmx = makefilename(".dmx");
    f=fopen(dmx,"wb"); if (!f) die("write %s: %s", dmx, strerror(errno));
    fwrite(encdata, 1, encsize, f);
    fclose(f);
    my_free(dmx);
  }
#endif

	// cheat a bit: don't do string compression for now
  if (compression == 2)
	compression = 1;

  fwrite(&compression, 1, 1, dat);

  switch (compression) {
	case 0:
		// write the flat data structure
		rle = encdata+16;	// not really RLE encoded!
		rlelen = encsize-16;
		fwrite(&rlelen, 4, 1, dat);
		fwrite(rle, 1, rlelen, dat);
		break;
	case 1:
		// rle-encode the rest of the file
		if (!LocoTool_rleencode(encdata+16, encsize-16, &rle, &rlelen))
			die("Error while RLE encoding");
		fwrite(&rlelen, 4, 1, dat);
		fwrite(rle, 1, rlelen, dat);
		my_free(rle);
		break;
	case 3:
		LocoTool_scramble(encdata+16, encsize-16);
		rlelen = encsize-16;
		fwrite(&rlelen, 4, 1, dat);
		fwrite(encdata+16, 1, encsize-16, dat);
		break;
	default:
		die("Can't encode chunk compression %d", compression);
  }

  fclose(dat);

  // make a backup of the old .dat if not .bak exists yet
  if (!access(datname, 0)) {
	if (access(bakname, 0)) {
		rename(datname, bakname);
	} else {
		remove(datname);
	}
  }
  rename(tmpname, datname);

  my_free(encdata);

  my_free(datname);
  my_free(tmpname);
  my_free(bakname);

  clearoptions();

  err_return = NULL;
  return 1;
}
